import random

from graphs.goal_graph import GoalGraph
from search_methods import bfs_search, dfs_search
from theory.prob_theory import expectation_truncated_geometric, prod


class CompleteTree(GoalGraph):
    """Generic class for complete tree goal graphs"""

    def __init__(self, goal_prob=[0, 0, 0.15, 0.1, 0],
                 branching_factor=2, create_graph=True):
        GoalGraph.__init__(self)
        self.D = len(goal_prob)-1
        self.goal_prob_vec = goal_prob

        def add_children(v, d, max_d):
            if goal_prob[d] > random.random():
                self.make_goal(v)
            if d == max_d:
                return  # add no more children
            c_list = self.add_vertex(branching_factor)  # add two chilren
            for c in c_list:
                self.add_edge(v, c)
                add_children(c, d+1, max_d)

        if create_graph:
            add_children(self.root, 0, self.D)

    def depth(self):
        return self.D

    def expected_bfs(self):
        """Expected BFS search time in a complete tree"""
        p = list(self.goal_prob_vec)
        p.append(1)
        f = [self.level_has_first_goal(k) for k in range(len(p))]
        return sum([f[k]*(2 ** k - 1 + expectation_truncated_geometric(pk, 2 ** k))
                    for k, pk in enumerate(p) if pk != 0])

    def expected_bfs_approx(self):
        """Expected BFS search time approximation in Prop 10"""
        p = list(self.goal_prob_vec)
        p.append(1)
        f = [self.level_has_first_goal(k) for k in range(len(p))]
        return sum([f[k]*(2**k + 1/pk)
                    for k, pk in enumerate(p) if pk != 0])

    def expected_bfs_g(self):
        return self.expected_bfs() - (1-self.has_goal())*2**(self.D+1)

    def level_has_goal(self, k):
        """The probability that level k has a goal"""
        p = self.goal_prob_vec
        if k < len(p):
            return 1-(1-p[k])**(2**k)
        else:
            return 1

    def has_goal(self):
        """The probability that a goal exists"""
        p = self.goal_prob_vec
        return 1 - prod([(1 - self.level_has_goal(k)) for k, pk in enumerate(p)])

    def level_has_first_goal(self, k):
        """The probability that level k has the first goal"""
        # p = self.goal_prob_vec
        return (self.level_has_goal(k) *
                prod([(1 - self.level_has_goal(j)) for j in range(k)]))

    def approx_expected_bfs(self):
        """Approximation of the more complicated formula for expected_bfs"""
        p = list(self.goal_prob_vec)
        p.append(1)
        return sum([self.level_has_first_goal(k)*(2**k + 1/pk)
                    for k, pk in enumerate(p) if pk != 0])

    def expected_dfs(self):
        p = self.goal_prob_vec
        # if max(p) == 1: return np.argmax(p)+1
        D = self.D
        # level rates
        l = [-math.log(1-pk) for pk in p]
        # total rate
        tr = sum([lk/(2**(D-k+1)) for (k, lk) in enumerate(l)])
        return 1/max(tr, 2**(-D))

    def predicted_difference(self):
        """DFS-BFS. Negative means DFS better"""
        return self.expected_dfs()-self.expected_bfs()

    def predicted_difference_g(self):
        """DFS-BFS. Negative means DFS better"""
        return self.expected_dfs()-self.expected_bfs_g()

    def predicted_log_ratio(self):
        """log(DFS/BFS). Negative means DFS better"""
        return np.log(self.expected_dfs()/self.expected_bfs())

    def prediction_correct(self):
        return ((self.expected_dfs() - self.expected_bfs()) *
                (dfs_search(self) - bfs_search(self)) > 0) * 1.0

    def bfs_better_theory(self):
        return self.expected_dfs() > self.expected_bfs()

    feature_list = [
        depth,
        GoalGraph.shallowest_goal,
        expected_bfs,
        # expected_dfs,
        bfs_search
    ]

    @staticmethod
    def random_init():
        depth = random.randint(5, 7)
        peak = depth-random.randint(1, 4)
        interval = random.randint(1, 4)
        goal_prob = [(peak-interval < i < peak+interval)*0.03/depth
                     for i in range(depth+1)]
        return {
            depth     : depth,
            goal_prob : goal_prob
        }