import graph_tool as gt
from graph_tool import topology, stats, draw

from search_methods import bfs_search, dfs_search
from theory.prob_theory import hist2mean, hist2stdev


class GoalGraph(gt.Graph):
    """Generic graph with goals"""

    def __init__(self, directed=False):
        gt.Graph.__init__(self, directed=directed)
        self.root = self.add_vertex()
        self.goal = self.new_vertex_property("bool")
        self.vertex_properties["goal"] = self.goal
        self.goals = list()

    def make_goal(self, v):
        if not self.vp.goal[v]:
            self.vp.goal[v] = True
            try:
                self.goals.append(v)
            except:
                self.goals = list()
                self.goals.append(v)

    # Feature Methods

    def number_of_goals(self):
        return len(self.goals)

    def goal_frequency(self):
        return self.number_of_goals()/self.num_vertices()

    def pseudo_diameter(self):
        return topology.pseudo_diameter(self)[0]

    def shallowest_goal(self):
        if len(self.goals) == 0:
            raise NameError('No goal')
        m = min([topology.shortest_distance(self, self.root, v,
                                            max_dist=self.num_vertices()+1)
                 for v in self.goals])
        if m > self.num_vertices():
            raise NameError('No goal reachable')
        else:
            return m

    def degree_histogram(self):
        return stats.vertex_hist(self, "out")

    def degree_mean(self):
        return hist2mean(self.degree_histogram())

    def degree_stdev(self):
        return hist2stdev(self.degree_histogram())

    def bfs_beats_dfs(self):
        return dfs_search(self) - bfs_search(self) > 0

    def bfs_wins(self):
        if dfs_search(self) > bfs_search(self):
            return "True"
        elif dfs_search(self) == bfs_search(self):
            return "Tie"
        else:
            return "False"

    def dfs_minus_bfs(self):
        return dfs_search(self) - bfs_search(self)

    def plotGraph(self, text=None):
        text = text or self.vertex_index
        pos = gt.draw.arf_layout(self)
        pos[self.vertex(0)] = [0.1, 0.1]
        pos = gt.draw.arf_layout(self, pos=pos)
        color = self.new_vertex_property("string")
        for v in self.vertices():
            if self.goal[v]:
                color[v] = 'g'
            else:
                color[v] = 'w'
        color[self.root] = 'b'
        draw.graph_draw(self, pos, vertex_color=color,
                        vertex_fill_color=color,
                        vertex_text=text,
                        vertex_font_size=10,
                        output_size=(2000, 2000))