import re

import numpy as np

from graphs.goal_graph import GoalGraph


class Grammar(GoalGraph):
    """Generic grammar graph"""

    def __init__(self,
                 start_string = "S",
                 goal_strings = ["aaa"],
                 rules        = [('S', ''), ('S','Sa'), ('S','Sb')],
                 max_depth    = 3, # negative = infinity
                 create_graph = True
                ):
        GoalGraph.__init__(self, directed=True)
        self.string       = self.new_vertex_property("string")
        self.node         = {} # dictionary string->node
        self.goal_strings = goal_strings
        self.rules        = rules
        self.max_depth    = max_depth

        def add_children(v, max_d):
            s = self.string[v]
            if s in goal_strings: self.make_goal(v)
            if len(self.remove_S(s)) == max_d:
                loc_rules = [('S', '')]
            else: loc_rules = rules
            for r in loc_rules:
                new_strings = self.apply_rule(r, s)
                for new_s in new_strings:
                    if new_s in self.node:
                        # node already exists, just add a link to it
                        self.add_edge(v, self.node[new_s])
                    else:
                        # create a node and recursively add its children
                        child = self.add_node(new_s)
                        self.add_edge(v, child)
                        add_children(child, max_d)

        self.string[self.root]  = start_string
        self.node[start_string] = self.root
        if create_graph:
            add_children(self.root, max_depth)

    def remove_S(self, string):
        """Return string with S removed if occurs"""
        ns = self.apply_rule(('S', ''), string)
        if len(ns) > 0:
            return ns[0]
        return string

    def apply_rule(self, rule, string):
        """Return the list of strings creatable from string and rule"""
        (o, n) = rule
        subs = [(i.start(), i.end()) for i in re.finditer(o, string)]
        new_s = set()
        for (a, b) in subs:
            new_s.add(string[:a] + n + string[b:])
        return list(new_s)

    def add_node(self, string):
        if string in self.node:  # already added
            return self.node[string]
        else:
            v = self.add_vertex()
            self.string[v] = string
            self.node[string] = v
            return v

    def number_of_rules(self):
        return len(self.rules)

    def maximum_depth(self):
        return self.max_depth

    def num_goal_strings(self):
        return len(self.goal_strings)

    def mean_len_goal_str(self):
        return np.mean([len(s) for s in self.goal_strings])

    def stdev_len_goal_str(self):
        return np.std([len(s) for s in self.goal_strings])

    def grammar_power(self):
        return self.number_of_goals()/self.num_goal_strings()

    def plotGraph(self):
        GoalGraph.plotGraph(self, text=self.string)