
# import timeit
import operator
from queue import Queue

import numpy as np

from graphs.goal_graph import GoalGraph
from n_puzzle import n_puzzle_params
from theory.prob_theory import discrete_gaussian

P = np.array([[0, 1, 0],  # corner
              [2/3, 0, 1/3],  # edge
              [0, 1, 0]])  # middle

Pn = P.dot(P).dot(P).dot(P).dot(P).dot(P).dot(P).dot(P).dot(P).dot(P)
Pstable = (Pn+Pn.dot(P))/2

type_distribution = Pstable[0, :]
naive_type_distribution = np.array([1, 4, 4])/9

t_dist = [None for i in range(32)]
t_dist[0] = np.array([1, 0, 0])  # adapt depending on type of starting state
for i in range(len(t_dist)-1):
    t_dist[i + 1] = t_dist[i].dot(P)
    assert sum(t_dist[i+1]) == 1


class State():

    def __init__(self, description):
        if isinstance(description, State):
            self.matrix = np.array(description.matrix)
            self.position = description.position
        else:
            n = int(np.sqrt(len(description)))
            self.matrix = np.reshape(description, (n, n))
            self.update_position()
        assert self.matrix[self.position] == 0

    def update_position(self):
        """position of 'empty' tile"""
        (x, y) = np.where(self.matrix == 0)
        self.position = (int(x), int(y))

    def equals(self, other):
        return np.array_equal(self.matrix, other.matrix)

    def __str__(self):
        return self.matrix.__str__()

    def __hash__(self):
        return tuple(map(tuple, self.matrix)).__hash__()


class NPuzzle():

    def __init__(self, n=3):
        self.actions = [(0, 1), (0, -1), (1, 0), (-1, 0)]
        self.n = n
        self.goal_state = self.generate_goal_state()
        self.start_state = self.generate_start_state()

    def generate_start_state(self):
        return State(np.random.permutation(self.n**2))

    def generate_goal_state(self):
        l = list(range(1, self.n**2))
        l.append(0)
        return State(l)

    def swap_elements(self, state, el1, el2):
        assert isinstance(state, State)
        assert isinstance(el1, tuple)
        tmp = state.matrix[el1]
        state.matrix[el1] = state.matrix[el2]
        state.matrix[el2] = tmp
        state.update_position()

    def available_actions(self, state):
        available = []
        new = []
        for a in self.actions:
            new_position = tuple(map(operator.add, state.position, a))
            if (0 <= new_position[0] <= self.n -1 and
                0 <= new_position[1] <= self.n -1):
                # if (0, 0) <= new_position < (self.n - 1, self.n - 1):
                available.append(a)
                new.append(new_position)
                # print("accepting", new_position)
            else:
                pass
                # print("rejecting", new_position)
        # print("available_actions", state.position, available, new)
        return available

    def take_action(self, state, a):
        state_copy = State(state)
        new_position = tuple(map(operator.add, state.position, a))
        self.swap_elements(state_copy, state.position, new_position)
        return state_copy

    def is_goal(self, state):
        return np.array_equal(state.matrix, self.goal_state.matrix)

    def misplaced_tiles(self, state):
        return (self.n**2 -
                len(np.where((state.matrix -
                              self.goal_state.matrix) == 0)[0]))

    def manhattan_distance(self, state=None):
        if state is None:
            state = self.start_state
        dist = 0
        for i in range(self.n**2):
            pos = np.array(np.where(state.matrix == i))
            gpos = np.array(np.where(self.goal_state.matrix == i))
            dist += int(np.linalg.norm((pos - gpos), ord=1))
        return dist


class NPuzzleGraph(GoalGraph):

    def __init__(self, n=3, from_goal=0, verbose=False):
        GoalGraph.__init__(self)
        self.verbose = verbose
        self.npuzzle = NPuzzle(n)
        if from_goal == 0:
            self.npuzzle.start_state = self.npuzzle.goal_state
        elif from_goal == 1:  # To start from edge
            self.npuzzle.start_state = self.npuzzle.goal_state
            self.npuzzle.swap_elements(self.npuzzle.start_state,
                                       (2, 2), (2, 1))
        elif from_goal == 2:  # To start from middle
            self.npuzzle.start_state = self.npuzzle.goal_state
            self.npuzzle.swap_elements(self.npuzzle.start_state,
                                       (2, 2), (1, 1))
        self.visited = {}
        self.state = self.new_vertex_property("string")
        self.vertex_properties["state"] = self.state  # internalise
        self.depth = self.new_vertex_property("int")
        self.vertex_properties["depth"] = self.depth  # internalise
        self.mt = self.new_vertex_property("int")  # misplaced tiles
        self.vertex_properties["mt"] = self.mt
        self.mh = self.new_vertex_property("int")  # manhattan dist
        self.vertex_properties["mh"] = self.mh
        self.vertex_properties["visit_count"] = self.new_vertex_property("int")
        self.goal_paths = []
        self.graph_index = {}
        self.remove_vertex(self.root)  # To manually set root
        self.root = self.visit(self.npuzzle.start_state, None)
        self.traverse_npuzzle(self.npuzzle)
        self.goal_node = self.node(self.npuzzle.goal_state)
        self.make_goal(self.goal_node)
        self.depth_dist = None

    def visit(self, state, parent):
        # visit_counter += 1
        if not self.been_visited(state):
            n = self.add_vertex()
            self.state[n] = state.__str__()
            self.graph_index[state.__str__()] = n
            if parent is not None:
                self.depth[n] = self.depth[self.node(parent)]+1
            self.mt[n] = self.npuzzle.misplaced_tiles(state)
            self.mh[n] = self.npuzzle.manhattan_distance(state)
        else:
            n = self.graph_index[state.__str__()]
        self.vp.visit_count[n] += 1
        self.visited[state.__str__()] = parent
        if parent is not None:
            m = self.graph_index[parent.__str__()]
            self.add_edge(m, n)
        return n

    def been_visited(self, state):
        return state.__str__() in self.visited

    def retrieve_path(self, state, path=None):
        if path is None:
            path = []
        path.insert(0, state)
        if state.equals(self.npuzzle.start_state):
            return path
        else:
            parent = self.visited[state.__str__()]
            return self.retrieve_path(parent, path)

    def node(self, state):
        return self.graph_index[state.__str__()]

    def traverse_npuzzle(self, npuzzle):
        """BFS"""
        Q = Queue()
        npuzzle.start_state.anc = {}
        Q.put(npuzzle.start_state)
        self.depth[self.node(npuzzle.start_state)] = 0
        if npuzzle.is_goal(npuzzle.start_state):
            self.goal_paths.append(
                len(self.retrieve_path(self.npuzzle.start_state)))
            self.make_goal(self.graph_index[
                self.npuzzle.start_state.__str__()])
        state_counter = 0
        new_state_counter = 0
        while (not Q.empty()) and new_state_counter < 188000:
            if self.verbose:
                print(Q.qsize(), state_counter,
                      new_state_counter, len(self.goal_paths))
            s = Q.get()
            for a in npuzzle.available_actions(s):
                state_counter += 1
                new_s = npuzzle.take_action(s, a)
                if self.been_visited(new_s):
                    self.add_edge(self.graph_index[s.__str__()],
                                  self.graph_index[new_s.__str__()])
                else:
                    Q.put(new_s)
                self.visit(new_s, s)
                if npuzzle.is_goal(new_s):
                    self.make_goal(self.graph_index[new_s.__str__()])
                    path = self.retrieve_path(new_s)
                    self.goal_paths.append(len(path))

        return self

    def depth_distribution(self):
        d = [0 for i in range(32)]
        for v in self.vertices():
            d[self.vp.depth[v]] += 1
        return d

    def mh_distributions(self):
        d = [[0 for i in range(32)] for i in range(32)]
        for v in self.vertices():
            d[self.vp.depth[v]][self.vp.mh[v]] += 1
        return d

    def nodes_of_depth(self):
        d = {}
        for v in self.vertices():
            if self.vp.depth[v] not in d:
                d[self.vp.depth[v]] = []
            d[self.vp.depth[v]].append(v)
        return d

    def nodes_of_mh(self):
        d = {}
        for v in self.vertices():
            if self.vp.mh[v] not in d:
                d[self.vp.mh[v]] = []
            d[self.vp.mh[v]].append(v)
        self.mh_nodes = d
        return d

    def nodes_of_mt(self):
        d = {}
        for v in self.vertices():
            if self.vp.mt[v] not in d:
                d[self.vp.mt[v]] = []
            d[self.vp.mt[v]].append(v)
        self.mt_nodes = d
        return d

    def goals_of_depth(self):  # TODO: How?
        d = {}  # mh -> (level -> # goals)
        for v in self.vertices():
            if self.vp.goal(v):
                if self.vp.depth[v] not in d:
                    d[self.vp.depth[v]] = 0
                d[self.vp.depth[v]] += 1
        self.goals_by_level = d
        return d

    def nodes_of_degree(self):
        types = {2: [], 3: [], 4: []}
        for v in self.vertices():
            types[len(set(v.out_neighbours()))].append(v)
        assert len(types[2]) == len(types[3]) == 4*len(types[4])
        return types

    def mh_to_depthdist(self):
        mh = NPuzzleGraph.nodes_of_mh(self)
        mh_to_d = {}
        for mhl in mh:
            mh_to_d[mhl] = np.array([0 for _ in range(32)])
            for v in mh[mhl]:
                mh_to_d[mhl][self.vp.depth[v]] += 1
        mh_to_d = {e: mh_to_d[e]/sum(mh_to_d[e]) for e in mh_to_d}
        return mh_to_d

    def local_branching_factors(self):
        """Average branching factor using equilibrium dist"""
        bdup = []
        bddown = []
        types = NPuzzleGraph.nodes_of_degree(self)
        for t in sorted(types.keys()):
            bdt = [[0, {-1: 0, 0: 0, 1: 0}] for i in range(32)]
            for v in types[t]:
                dv = self.vp.depth[v]
                bdt[dv][0] += 1
                for w in set(v.out_neighbours()):
                    dw = self.vp.depth[w]
                    bdt[dv][1][dw-dv] += 1
            bd1 = [bdt[d][1][-1]/bdt[d][0]
                   if bdt[d][0] > 0 else bdt[d][1][-1]
                   for d in range(len(bdt)-1)]
            bd2 = [bdt[d][1][1]/bdt[d][0]
                   if bdt[d][0] > 0 else bdt[d][1][1]
                   for d in range(len(bdt)-1)]
            bdup.append(bd1)
            bddown.append(bd2)
        avup = [t_dist[i].dot((bdup[0][i], bdup[1][i], bdup[2][i]))
                for i in range(31)]
        avdown = [t_dist[i].dot((bddown[0][i], bddown[1][i], bddown[2][i]))
                  for i in range(31)]
        return avup, avdown

    def global_branching_factors(self):
        dd = NPuzzleGraph.depth_distribution(self)
        return [dd[i+1]/dd[i] for i in range(len(dd)-1)]

    @staticmethod
    def goal_distribution(mh, bgdown=n_puzzle_params.bgdown, D=20):
        """Estimated goal distribution from Manhattan distance"""
        sd = 3.5
        mu = 1.5
        number_of_goals = 1
        pv = np.array([number_of_goals*discrete_gaussian(i, mu=mu*mh, sd=sd)/bgdown**i
                       for i in range(D+1)])
        return pv.clip(0, 1)

    def true_goal_distribution(self, depth=None, D=20):
        """Single goal level distribution given depth of goal is known"""
        if not depth:
            depth = self.vp.depth[self.root]
        if not self.depth_dist:
            self.depth_dist = NPuzzleGraph.depth_distribution(self)
        pv = [0 for _ in range(D + 1)]
        if depth < len(pv):
            pv[depth] = 1 / self.depth_dist[depth]
        return pv


# TODO: Probably never used. Remove?
def analyze_goal_levels(n=3):
    G = NPuzzleGraph(n)
    d = {}  # (mt, mh) -> array([gl1, ..., gl31])
    for i in G.vertices():
        (mh, mt, depth) = (G.vp.mh[i], G.vp.mt[i], G.vp.depth[i])
        if (mh, mt) not in d:
            d[(mh, mt)] = np.zeros(32)
        d[(mh, mt)][depth] += 1
    for i in d:
        print(i[0], i[1], ', '.join(map(lambda x: str(int(x)), d[i])),
              sep=", ")

# TODO: Probably never used. Remove?
def analyze_desc_count(n=3, from_goal=0):
    G = NPuzzleGraph(n, from_goal=from_goal)
    d = {l: 0 for l in range(32)}  # level -> # nodes on level
    for i in G.vertices():
        d[G.vp.depth[i]] += 1
    for i in d:
        print(i, d[i])