import math

import numpy as np
import prob_theory
from scipy.special import binom


class Branch2Counters():

    def __init__(self,
                 blup=None, bldown=None, bgup=None, bgdown=None, max_depth=None):
        self.blup = blup
        self.bldown = bldown
        self.bgup = bgup
        self.bgdown = bgdown

        self.max_depth = max_depth

        # self.pupup = blup / bldown
        # self.pupdown = 1 - self.pupup
        # self.pdowndown = bgdown / bldown
        # self.pdownup = 1 - self.pdowndown

        b = blup + bldown - 1
        p_upup = (blup-1)/b
        p_downdown = (bldown-1)/b
        p_updown = bldown/b
        p_downup = blup/b

        #P = np.array([[pupup, pupdown], [pdownup, pdowndown]])
        P = np.array([[p_downup, p_downdown], [p_upup, p_updown]])
        Pstable = P.dot(P).dot(P).dot(P).dot(P).dot(P).dot(P).dot(P).dot(P).dot(P)
        Pstable = Pstable.dot(Pstable).dot(Pstable).dot(Pstable).dot(Pstable)
        self.depth_dist = Pstable[0, :]

    def K(self, n, d, r, fast=False):
        """Depth-to-depth K(l, d, r) in the paper"""
        dn = abs(d - n)
        rextra = r - dn
        rextra2 = math.floor(rextra/2)
        if rextra < 0:
            return 0
        if d > n:
            bdir = self.bgdown
        else:
            bdir = self.bgup
        s = 0
        if fast:
            return min(self.bgdown**d, \
                       bdir**dn*(1 + (dn + rextra2)*(self.bg*self.bup)**rextra2 *\
                                 (self.bldown-1)*(self.blup-1)))
        else:
            for m in range(1, rextra2+1):
                for t in range(1, m):
                    s += prob_theory.nCk(m - 1, t - 1) * prob_theory.nCk(dn + m, t) *\
                         bdir**dn *(self.bgdown*self.bgup)**rextra2 *((self.bldown-1)*(self.blup-1))**t
            return min(self.bgdown**d, s + bdir**dn)

    def Kbar(self, n, d, r, fast=False):
        """Depth-to-depth \bar K(l, d, r) in the paper"""
        bldown = self.bldown
        blup = self.blup
        dn = abs(d - n)
        rextra = r - dn
        rextra2 = math.floor(rextra/2)
        if rextra < 0:
            return 0
        if d > n:
            bdir = bldown
        else:
            bdir = blup
        s = 0
        if fast:
            return bdir**dn*(1 + (dn + rextra2) * (bldown * blup) ** rextra2 * (bldown - 1) * (blup - 1))
        else:
            for m in range(1, rextra2+1):
                for t in range(1, m):
                    s += prob_theory.nCk(m - 1, t - 1) * prob_theory.nCk(dn + m, t) *\
                         bdir**dn * (bldown * blup) ** rextra2 * ((bldown - 1) * (blup - 1)) ** t
            return s + bdir**dn

    def _depth_distribution(self, path_length, depth):
        """Probability P_n(d) at being at depth after path_length"""
        if (path_length - depth) % 2 == 1:
            return 0
        p = self.depth_dist[1]
        up = int(1/2*(path_length - depth))
        down = path_length - up
        return binom(path_length, down) * p**down * (1-p)**up

    def depth_distribution(self, path_length):
        return np.array([self._depth_distribution(path_length, k)
                         for k in range(self.max_depth)])

    def expected_depth(self, path_length):
        return self.depth_distribution(path_length).dot(list(range(self.max_depth)))

    def L(self, D, l, d):
        """Length-to-depth counter L(n, d)"""
        r = D - l
        return sum([self._depth_distribution(l, n)*self.K(n, d, r)
                    for n in range(l+1)])

    def Lbar(self, D, l, d):
        """Non-unique length-to-depth counter \bar L(n, d)"""
        r = D - l
        return sum([self._depth_distribution(l, n)*self.Kbar(n, d, r)
                    for n in range(l+1)])

