import pickle
import random
import sys

import graph_tool
import numpy as np
import pandas
from matplotlib import pyplot as plt

from graphs.n_puzzle import NPuzzleGraph
from n_puzzle.misc_methods import print_vectors
from search_methods import DFS_tree, BFS_graph
from theory import counters2searchtime


def load_graph(filename, make_class):
    G = graph_tool.load_graph(filename)
    return G


def search_time(G, searcher, D=20, iterations=10, verbose=False):
    """
    Generate data frame with search times for supplied search method
    for different Manhattan distances.
    Use plot_search_time function to visualise.
    """
    data = []
    mh_nodes = NPuzzleGraph.nodes_of_mh(G)
    for mh in mh_nodes:
        node_list = mh_nodes[mh]
        for i in range(iterations):
            v = random.choice(node_list)
            G.root = v
            depth = G.vp.depth[v]
            (s_time, visited, goal_found) = searcher(G, max_path_length=D, verbose=verbose)
            data.append([depth, mh, s_time, visited, goal_found])
            if verbose:
                print(mh, i, data[-1], file=sys.stderr)
    return pandas.DataFrame(data,
                            columns=("depth", "mh", "emp", "visited", "goal found"))


def search_time_all(G, iterations=10, verbose=False):
    bfs_df = search_time(G, BFS_graph, iterations=iterations, verbose=verbose)
    print("BFS done")
    dfs_df = search_time(G, DFS_tree, iterations=iterations, verbose=verbose)
    print("DFS done")
    # dfsg_df = iterate_DFSg(G, iterations=iterations)
    return {'BFS': bfs_df, 'DFSt': dfs_df}   # , 'DFSg': dfsg_df}


markers = ['+', '+', '.', 'o', 'o', '*', '*', ',', ',', '-']


def plot_search_times(dic, L=None, Lbar=None, goal_prob=None, D=20, filename=None, allow_no_goal=True, log_plot=False):
    """
    Expects a data from with columns 'emp', and 'mh'
    for empirical outcome and Manhattan distance.
    Compares emp with tree search bounds.
    """
    data = {}

    # Empirical values
    for d in dic:
        df = dic[d]
        mhs = sorted(set(df['mh']))
        data['mh'] = mhs
        if allow_no_goal:
            means = [df.loc[df['mh'] == i]['emp'].mean() for i in mhs]
        else:
            means = [df.loc[df['mh'] == i].loc[df['goal found']]['emp'].mean() for i in mhs]
        data[d+"mean"] = means
        #plt.plot(df['mh'], df['emp'], ",", label="emp "+d)
        plt.plot(mhs, means, label="means "+d)
        #        mdic[d] = means

    gd = goal_prob

    if gd and L and Lbar:
        # DFS tree search
        upper = [counters2searchtime.upper_tdfs(D, gd(mh, D=D), L, Le=Lbar, allow_no_goal=allow_no_goal)
                 for mh in mhs]
        lower = [counters2searchtime.lower_tdfs(D, gd(mh, D=D), L, Le=Lbar, allow_no_goal=allow_no_goal)
                 for mh in mhs]
        data["DFStlower"] = lower
        data["DFStupper"] = upper
        plt.plot(mhs, upper, "--", label="DFS upper")
        plt.plot(mhs, lower, "--", label="DFS lower")

        # # DFS graph search
        # gupper = [grammar.upper_tdfs(D, gd(mh, D=D), K, allow_no_goal=allow_no_goal) for mh in mhs]
        # glower = [grammar.lower_tdfs(D, gd(mh, D=D), K, allow_no_goal=allow_no_goal) for mh in mhs]
        # data["DFSglower"] = glower
        # data["DFSgupper"] = gupper
        # plt.plot(mhs, gupper, "--", label="graph upper")
        # plt.plot(mhs, glower, "--", label="graph lower")

        # BFS
        bfs = [counters2searchtime.tbfs(D, gd(mh, D=D), L, allow_no_goal=allow_no_goal)
               for mh in mhs]
        data["BFStheory"] = bfs
        plt.plot(mhs, bfs, "--", label="BFS theory")

    if filename:  # print to file
        print_vectors(filename, data)
    else:  # display plot
        #plt.yscale('log')
        # plt.axis((1, 25, 1, 10**6))
        plt.legend(loc=0)
        plt.xlabel("Manhattan distance")
        plt.ylabel("Search time")
        if log_plot:
            plt.yscale('log')
        plt.show()

    return data


def plot_data(data, xcol="mh"):
    x = data[xcol]
    for col in data:
        y = data[col]
        if "upper" in col or "lower" in col or "theory" in col:
            plt.plot(x[:17], y[:17], "--", label=col)
        else:
            plt.plot(x[:17], y[:17], label=col)
    # plt.yscale('log')
    plt.legend()
    plt.show()


def scatter_plot(G, D=20, iterations=100):
    data_dfs = {"bfs_win": [], "gp": [], "mu": []}
    data_bfs = {"bfs_win": [], "gp": [], "mu": []}
    # gps = [0.0001, 0.0003, 0.001, 0.003, 0.01, 0.03, 0.1, 0.3]
    for i in range(iterations):
        gp = 10**(-np.random.uniform(0, 4))
        mu = np.random.uniform(1, D)
        NPuzzleGraph.add_sgl(G, mu, gp=gp)
        (dfs_time, _, _) = DFS_tree(G, max_path_length=D)
        (bfs_time, _, _) = BFS_graph(G, max_path_length=D)
        if bfs_time < dfs_time:
            data_bfs["mu"].append(mu)
            data_bfs["gp"].append(gp)
            data_bfs["bfs_win"].append(bfs_time < dfs_time)
        else:
            data_dfs["mu"].append(mu)
            data_dfs["gp"].append(gp)
            data_dfs["bfs_win"].append(bfs_time < dfs_time)
        print(i)
    print_vectors("8-puzzle-scatter-bfs.dat", data_bfs)
    print_vectors("8-puzzle-scatter-dfs.dat", data_dfs)


def decision_boundary(goal_distribution, b2g, D=20):
    # global sd, mu, gd
    global tdata
    tdata = {"bfs_win": [], "gp": [], "mu": []}
    gps = [0.0001, 0.0003, 0.001, 0.003, 0.01, 0.03, 0.1, 0.3, 0.95]
    mus = list(range(1, D+1))
    f = open("8-puzzle-decision_boundary.dat", "w", 1)
    print("gp", "mu", "diff", sep="\t", file=f)
    for gp in gps:
        print(file=f)
        for mu in mus:
            gd = goal_distribution(mu, gp=gp)
            tbfs = counters2searchtime.tbfs(D, gd, b2g.L, allow_no_goal=True)
            tdfs = counters2searchtime.average_tdfs(D, gd, b2g.L, Le=b2g.Lbar, allow_no_goal=True)
            tdata["mu"].append(mu)
            tdata["gp"].append(gp)
            tdata["bfs_win"].append(tbfs - tdfs)
            print(gp, mu)
            print(gp, mu, tbfs - tdfs, sep="\t", file=f)
    f.close()
    # print_vectors("8-puzzle-decision_boundary.dat", tdata)


if __name__ == "__main__":
    G = graph_tool.load_graph("new_data/n3graph.xml.gz")
    runtimes = search_time_all(G, iterations=100, verbose=True)
    pickle.dump(runtimes, open("new_data/runtimes.p", "wb"))
