import random
import sys
from queue import Queue

from graph_tool import search


def DFS_tree(G, max_path_length=20, verbose=True):
    expanded = 0
    visited = set()

    def dfs_rec(v, path):
        nonlocal G, max_path_length, expanded, visited
        if G.vp.goal[v]:
            return True

        # Recurse
        expanded += 1
        visited.add(v)
        if len(path) < max_path_length:
            path.add(v)
            children = list(set(v.out_neighbours()) - path)
            random.shuffle(children)
            for c in children:
                ret = dfs_rec(c, path=set(path))
                if ret:
                    return ret

    ret = dfs_rec(G.root, path=set())
    if ret:
        return (expanded, len(visited), True)
    else:
        if verbose:
            print("no goal found", file=sys.stderr)
        return (expanded + 1, len(visited), False)


def DFS_graph(G, max_path_length=20, verbose=True):
    expanded = 0
    visited = set()
    G.distance = G.new_vertex_property("int")
    G.distance[G.root] = 1

    def dfs_rec(v, path):
        nonlocal G, max_path_length, expanded, visited
        if G.vp.goal[v]:
            return True

        # Recurse
        expanded += 1
        visited.add(v)
        G.distance[v] = len(path) + 1
        if len(path) < max_path_length:
            path.add(v)
            children = list(set(v.out_neighbours()))
            random.shuffle(children)
            for c in children:
                if G.distance[c] == 0 or G.distance[c] > len(path) + 1:
                    ret = dfs_rec(c, path=set(path))
                    if ret:
                        return ret

    ret = dfs_rec(G.root, path=set())
    if ret:
        return (expanded, len(visited), True)
    else:
        if verbose:
            print("no goal found", file=sys.stderr)
        return (expanded + 1, len(visited), False)


def BFS_graph(G, max_path_length=20, verbose=True):
    depth = {}
    Q = Queue()
    Q.put(G.root)
    depth[G.root] = 0
    if G.vp.goal[G.root]:
        # print("root goal")
        return (len(depth), len(depth), True)
    while not Q.empty():
        v = Q.get()
        for new_s in set(v.out_neighbours()):
            if new_s not in depth:
                depth[new_s] = depth[v] + 1
                if depth[new_s] < max_path_length:
                    Q.put(new_s)
                if G.vp.goal[new_s]:
                    # print("goal found")
                    return (len(depth), len(depth), True)
    if verbose:
        print("no goal found", file=sys.stderr)
    return (len(depth)+1, len(depth)+1, False)


def print_path(path):
    print("\nPath length", len(path), "\n")
    for s in path:
        print(s, "\n")


def bfs_search(graph):
    """Return the number of BFS visited nodes until goal found"""

    class MyBFSVisitor(search.BFSVisitor):
        discovered_counter = 0
        found_goal = False

        def discover_vertex(self, v):
            self.discovered_counter += 1
            # print("discovered",v)
            if graph.goal[v]:
                self.found_goal = True
                # print("found goal", v,"in", self.discovered_counter, "steps")
                raise search.StopSearch()

    visitor = MyBFSVisitor()
    search.bfs_search(graph, graph.root, visitor=visitor)
    if visitor.found_goal:
        return visitor.discovered_counter
    else:  # Search failed
        # return visitor.discovered_counter + 1
        raise NameError('No goal reached')


def dfs_search(graph):
    """Return the number of DFS visited nodes until goal found"""

    class MyDFSVisitor(search.DFSVisitor):
        discovered_counter = 0
        found_goal = False

        def discover_vertex(self, v):
            # print(v)
            self.discovered_counter += 1
            # print("discovered",v)
            if graph.goal[v]:
                self.found_goal = True
                # print("found goal", v,"in", self.discovered_counter, "steps")
                raise search.StopSearch()

    visitor = MyDFSVisitor()
    search.dfs_search(graph, graph.root, visitor=visitor)
    if visitor.found_goal:
        return visitor.discovered_counter
    else:  # Search failed
        # return visitor.discovered_counter + 1
        raise NameError('No goal reached')


def my_dfs_rec(graph, max_depth=20):
    """Alternative DFS implementation.
    Explore in same order as dfs_search"""
    discovered = {graph.root.__int__()}

    def rec(u, discovered, depth=0):
        discovered.add(u.__int__())
        print(discovered)
        if graph.goal[u]:
            return len(discovered)
        neigh = list(set(u.out_neighbours()))
        random.shuffle(neigh)
        if depth < max_depth:
            for v in neigh:
                print("considering", v)
                if not v.__int__() in discovered:
                    c = rec(v, discovered, depth=depth+1)
                    if c:
                        return c
    return rec(graph.root, discovered, depth=0)


def my_dfs(graph):
    """Alternative DFS implementation.
    Explore in different order than dfs_search"""
    stack = [graph.root]
    discovered = {graph.root.__int__()}  # set of discovered nodes
    counter = 0
    while len(stack) > 0:
        print(discovered)
        counter += 1
        u = stack.pop()
        if graph.goal[u]:
            return counter
        for v in set(u.out_neighbours()):
            if not v.__int__() in discovered:
                if graph.goal[v]:
                    return counter
                discovered.add(v.__int__())
                stack.append(v)