#
# Author:  Tom Everitt
# Email:   tom4everitt@gmail.com
# website: http://tomeveritt.se
#
# License:
# Use as you like as long as you credit me in publications.
#

######################################################
#
# This file has methods computing theoretical search time estimates
# from length-to-depth counters and goal distributions
#

import numpy as np

from theory.prob_theory import exp_trunc_geo, prod


lvd = {}
def lv(D, L):
    """
    Matrix (D+1)x(D+1), element (n,d) is the number of elements on level d
    reachable from a^n given function L(D,n,d)
    """
    try:
        return lvd[(D, L)]
    except:
        lvd[(D, L)] = np.array([[L(D, n, d) for d in range(D+1)] for n in range(D+1)])
        return lvd[(D, L)]


def sv(D, L):
    """Vector where element n is the size of subtree under a^n"""
    return sum(lv(D, L).T)


def tv(D, L):
    """Vector where element n is the number of new nodes seen from a^n"""
    s = sv(D, L)
    return np.append(s[:-1] - s[1:], L(D, D, D))


def av(D, L):
    """Matrix where element (n,d) is the new nodes seen on level d from a^n"""
    lastrow = [0]*(D+1)
    lastrow[D] = L(D, D, D)
    return np.append(lv(D, L)[:-1]-lv(D, L)[1:], [lastrow], axis=0)


def lmbdav(D, p, L):
    """Vector of length D+1 where element n is the probability
    of finding goal under a^n"""
    a = av(D, L)
    a[a < 0] = 0
    assert(np.min(a) >= 0)
    return [1 - prod([(1-p[d])**a[n, d] for d in range(0, D+1)])
            for n in range(D+1)]


def prob_first_goalv(D, p, L):
    """Element n is the probability that the first goal is under a^n"""
    l = lmbdav(D, p, L)
    assert(min(l) >= 0)
    p = [l[n]*prod([(1-lk) for lk in l[n+1:D+1]]) for n in range(D+1)]
    if sum(p) > 0:
        return p/sum(p)
    else:
        return p


def lower_tdfs(D, p, L, Le=None, allow_no_goal=True):
    """Lower bound on expected DFS search time"""
    assert D+1 == len(p)
    if Le is None:
        Le = L
    f = list(prob_first_goalv(D, p, L))
    f.insert(0, 1-sum(f))  # \phi_{-1}, to account for no goal found
    f = np.array(f)
    s = list(sv(D, Le))
    s.append(0)  # S_{D+1}
    s = np.array(s)
    # return sum([s[n+1]*f[n] for n in range(0,D)])
    if allow_no_goal:
        return np.dot(s, f)
    elif sum(f[1:]) == 0:
        print("Warning: probability 0")
    return np.dot(s[1:], f[1:]/sum(f[1:]))


def upper_tdfs(D, p, L, Le=None, allow_no_goal=True):
    """Upper bound on expected DFS search time"""
    if Le is None:
        Le = L
    assert D+1 == len(p)
    f = list(prob_first_goalv(D, p, L))
    f.insert(0, 1-sum(f))  # To account for no goal found
    f = np.array(f)
    s = list(sv(D, Le))
    s.insert(0, s[0]+1)
    s = np.array(s)
    # return(f, s)
    # return sum([s[n]*f[n] for n in range(0,D+1)])
    if allow_no_goal:
        return np.dot(s, f)
    elif sum(f[1:]) == 0:
        print("Warning: probability 0")
    return np.dot(s[1:], f[1:]/sum(f[1:]))


def average_tdfs(D, p, L, Le=None, allow_no_goal=True):
    """Arithmetic mean of lower and upper bound"""
    lower = lower_tdfs(D, p, L, Le=Le, allow_no_goal=allow_no_goal)
    upper = upper_tdfs(D, p, L, Le=Le, allow_no_goal=allow_no_goal)
    return (lower + upper)/2


def tbfs(D, p, L, allow_no_goal=True):  # TODO: implement allow_no_goal
    """Expected BFS search time"""
    l = lv(D+1, L)
    # if extra_level:
    #    p = [p[k]/2**k for k in range(len(p))]
    p = list(p)
    p.append(1)
    # probability level has a goal
    g = np.array([1-(1-pk)**l[0, k] for (k, pk) in enumerate(p)])
    # probability level has the first goal
    f = np.array([g[n]*prod(1 - g[:n]) for n in range(D+2)])
    assert 0.99 <= sum(f) <= 1.01
    s = [sum(l[0, :n]) for n in range(D+2)]
    # expected search time on level n, given goal on level n
    lt = np.array([s[n] + exp_trunc_geo(p[n], l[0, n]) for n in range(D + 2)])
    if allow_no_goal:
        return np.dot(f, lt)
    else:
        return np.dot(f[:-1]/sum(f[:-1]), lt[:-1])