import math
import random

import numpy as np
from matplotlib import pyplot as plt
from scipy.special import comb

norm_factors = {}


def discrete_gaussian(x, mu=0, sd=3.5):

    def dens(x):
        """un-normalised density"""
        return math.exp(-(x - mu)**2/(2*sd**2))

    if sd in norm_factors:
        norm = norm_factors[sd]
    else:
        norm = 1/sum([dens(x) for x in range(int(mu-100*sd), int(mu+100*sd))])
        norm_factors[sd] = norm
    return norm*dens(x)


def prod(list):
    """Return the product of elements in received list of numbers"""
    p = 1
    for n in list:
        p = p*n
    return p


def expectation_truncated_geometric(p, m):
    """Let X ~ Geo(p), i.e. P(X=k)=(1-p)^(k-1)*p. Return E[ X | X <= m]."""
    if p < 0.00000001:
        return m/2 + 1  # approximate to avoid division by zero
    return (1 - (1-p)**m*(p*m+1))/(p*(1-(1-p)**m))


def sample_k(max):
    """Sample from cut-off 1/k distribution"""
    accept = False
    while not accept:
        k = random.randint(1, max+1)
        accept = random.random() < 1.0/k
    return k


def hist2mean(hist):
    """Compute mean from histogram on form (values, bucket_means)"""
    (values, bucket_means) = hist
    s = 0.0
    for i in range(len(values)):
        s += values[i]*bucket_means[i]
    return s/sum(values)


def hist2stdev(hist):
    """Compute stdev from histogram on form (values, bucket_means)"""
    (values, bucket_means) = hist
    mean = hist2mean(hist)
    sum_of_squares = 0
    for i in range(len(values)):
        sum_of_squares += values[i]*(bucket_means[i]**2)
    variance = sum_of_squares/sum(values) - mean**2
    return(math.sqrt(variance))


def random_string(alphabet, length):
    """Return random string from alphabet"""
    return ''.join([random.choice(alphabet) for i in range(length)])


def gamma(p, g):
    """gamma term in SGL decision boundary"""
    return np.log2(expectation_truncated_geometric(p, 2**g)-1)/2



def average(l):
    """Not sure if ever used"""
    return sum([i*l[i] for i, _ in enumerate(l)])/sum(l)


def exp_trunc_geo(p, m):
    """
    Truncated geometric expectation.
    Let X ~ Geo(p), i.e. P(X=k)=(1-p)^(k-1)*p. Return E[X | X <= m].
    """
    assert 0 <= p <= 1
    if p < 0.00000001:
        return m/2 + 1  # approximate to avoid division by zero
    return (1 - (1-p)**m*(p*m+1))/(p*(1-(1-p)**m))


def nCk(n, k):
    """ n choose k """
    assert n >= k
    assert k >= 0
    return int(np.round(comb(n, k, exact=False)))


def mean(X):
    s = 0
    for x in range(len(X)):
        s += x*X[x]
    return s


def nmean(X):
    s = 0
    for x in range(len(X)):
        s += x*X[x]/sum(X)
    return s


def variance(X):
    m = mean(X)
    v = 0
    for x in range(len(X)):
        v += (x-m)**2 * X[x]
    return v


def nvariance(X):
    m = nmean(X)
    v = 0
    for x in range(len(X)):
        v += (x-m)**2 * X[x]/sum(X)
    return v


def std(X):
    return np.sqrt(variance(X))


def nstd(X):
    return np.sqrt(nvariance(X))


def plot_means_and_std(dists, leg_prefix="", mark="-", legend_pos=0, show=True, filename=None):
    means = {}
    stdevs = {}
    for d in dists:
        means[d] = nmean(dists[d])
        stdevs[d] = nstd(dists[d])
    xx = np.array(sorted(list(means.keys())))
    ymeans = np.array([means[x] for x in xx])
    ystd = np.array([stdevs[x] for x in xx])
    yupper = ymeans + ystd
    ylower = ymeans - ystd
    if filename:
        f = open(filename, "w")
        print("index", "means", "upper", "lower", sep="\t", file=f)
        for i in range(len(xx)):
            print(xx[i], ymeans[i], yupper[i], ylower[i], sep="\t", file=f)
        f.close()
    plt.plot(xx, ymeans, mark, label=leg_prefix+"mean")
    plt.plot(xx, yupper,  mark, label=leg_prefix+"upper")
    plt.plot(xx, ylower,  mark, label=leg_prefix+"lower")
    if show:
        plt.legend(loc=legend_pos)
        plt.show()