import re
import csv
import math
import numpy as np
import random
import pandas
import matplotlib.pyplot as plt
import graph_tool as gt
from graph_tool import draw, generation, search, stats, topology


def prod(list):
    p = 1
    for n in list:
        p = p*n
    return p

def expectation_truncated_geometric(p, m):
    """Let X ~ Geo(p), i.e. P(X=k)=(1-p)^(k-1)*p. Return E[ X | X <= m]."""
    if p < 0.00000001: return m/2 + 1 # approximate to avoid division by zero
    return (1 - (1-p)**m*(p*m+1) )/(p*(1-(1-p)**m))

def functionalizer(g, x, bounds=(-100000, 100000)):
    (l,u) = bounds
    if not ( g(x,l) < 0 < g(x,u) and l<u  ): raise NameError("illegal input functionalizer")
    while math.abs(g(x,l)) < 0.000001:
        if g(x, (u-l)/2) < 0:
            l = (u-l)/2
        else:
            u = (u-l)/2
    return l

# Sample from cut-off 1/k distribution
def sample_k(max):
    accept = False
    while not accept:
        k = random.randint(1,max+1)
        accept = random.random() < 1.0/k
    return k

# Assumes histogram on form (values, bucket_means)
def hist2mean(hist):
    (values, bucket_means) = hist
    s = 0.0
    for i in range(len(values)):
        s += values[i]*bucket_means[i]
    return s/sum(values)

# Slightly off for small samples compared to statistics.stdev
# Source: http://mathforum.org/library/drmath/view/52199.html
def hist2stdev(hist):
    (values, bucket_means) = hist
    mean = hist2mean(hist)
    sum_of_squares = 0
    for i in range(len(values)):
        sum_of_squares += values[i]*(bucket_means[i]**2)
    variance = sum_of_squares/sum(values) - mean**2
    return(math.sqrt(variance))

def random_string(alphabet, length):
    """Return random string from alphabet"""
    return ''.join([random.choice(alphabet) for i in range(length)])


# Search algorithms

def bfs_search(graph):
    """Return the number of visited nodes."""
    
    class MyBFSVisitor(search.BFSVisitor):
        discovered_counter = 0
        found_goal = False
        def discover_vertex(self, v):
            self.discovered_counter += 1
            #print("discovered",v)
            if graph.goal[v]:
                self.found_goal = True
                #print("found goal", v,"in", self.discovered_counter, "steps")
                raise search.StopSearch()
    
    visitor = MyBFSVisitor()
    search.bfs_search(graph, graph.root, visitor=visitor)
    if visitor.found_goal:
        return visitor.discovered_counter
    else: # Search failed
        raise NameError('No goal reached')

def dfs_search(graph):
    """Return the number of visited nodes."""
    
    class MyDFSVisitor(search.DFSVisitor):
        discovered_counter = 0
        found_goal = False
        def discover_vertex(self, v):
            self.discovered_counter += 1
            #print("discovered",v)
            if graph.goal[v]:
                self.found_goal = True
                #print("found goal", v,"in", self.discovered_counter, "steps")
                raise search.StopSearch()
    
    visitor = MyDFSVisitor()
    search.dfs_search(graph, graph.root, visitor=visitor)
    if visitor.found_goal:
        return visitor.discovered_counter
    else: # Search failed
        raise NameError('No goal reached')



# Graph classes

class GoalGraph(gt.Graph):

    def __init__(self, directed=False):
        gt.Graph.__init__(self, directed=directed)
        self.root = self.add_vertex()
        self.goal = self.new_vertex_property("bool")
        self.goals = list()
    
    def make_goal(self, v):
        if not self.goal[v]:
            self.goal[v] = True
            self.goals.append(v)
    
    # Feature Methods
    
    def number_of_goals(self):
        return len(self.goals)
    
    def goal_frequency(self):
        return self.number_of_goals()/self.num_vertices()
    
    def pseudo_diameter(self):
        return topology.pseudo_diameter(self)[0]

    def shallowest_goal(self): 
        if len(self.goals) == 0: raise NameError('No goal')
        m = min([topology.shortest_distance(self, self.root, v, 
                                            max_dist=self.num_vertices()+1)
                 for v in self.goals])
        if m > self.num_vertices(): 
            raise NameError('No goal reachable') 
        else: return m

    def degree_histogram(self):
        return stats.vertex_hist(self, "out")
            
    def degree_mean(self):
        return hist2mean(self.degree_histogram())

    def degree_stdev(self):
        return hist2stdev(self.degree_histogram())
    
    def bfs_beats_dfs(self):
        return dfs_search(self) - bfs_search(self) > 0
        
    def plotGraph(self, text=None):
        text = text or self.vertex_index
        pos = gt.draw.arf_layout(self)
        pos[self.vertex(0)] = [0.1,0.1]
        pos = gt.draw.arf_layout(self, pos=pos)
        color = self.new_vertex_property("string")
        for v in self.vertices():
            if self.goal[v]:
                color[v] = 'g'
            else:
                color[v] = 'w'
        color[self.root] = 'b'
        draw.graph_draw(self, pos, vertex_color=color, 
                        vertex_fill_color=color, 
                        vertex_text=text, 
                        vertex_font_size=10,
                        output_size=(2000, 2000)
                    )


class Grammar(GoalGraph):

    def __init__(self, 
                 start_string = "a", 
                 goal_strings = ["c"], 
                 rules        = [('a','b'), ('b','c')], 
                 max_depth    = 20 # negative = infinity
                ): 
        GoalGraph.__init__(self, directed=True)
        self.string       = self.new_vertex_property("string")
        self.node         = {} # dictionary string->node
        self.goal_strings = goal_strings
        self.rules        = rules
        self.max_depth    = max_depth
        
        def add_children(v, d, max_d):
            s = self.string[v]
            if s in goal_strings: self.make_goal(v)
            if d == max_d: return
            for (o, n) in rules:
                subs = [(i.start(), i.end()) for i in re.finditer(o, s)]
                #print("found ", subs, "for",o,"in",s)
                for (a,b) in subs:
                    new_s = s[:a] + n + s[b:]
                    #print("new string", new_s)
                    if new_s in self.node:
                        # node already exists, just add a link to it
                        self.add_edge(v, self.node[new_s])
                    else:
                        # create a node and recursively add its children
                        child = self.add_node(new_s)
                        self.add_edge(v, child)
                        add_children(child, d+1, max_d)
                        
        self.string[self.root]  = start_string
        self.node[start_string] = self.root
        add_children(self.root, 0, max_depth)
            
    def add_node(self, string):
        if string in self.node: # already added
            return self.node[string]
        else:
            v = self.add_vertex()
            self.string[v]    = string
            self.node[string] = v
            return v
    
    def number_of_rules(self):
        return len(self.rules)
    
    def maximum_depth(self):
        return self.max_depth

    def number_of_goal_strings(self):
        return len(self.goal_strings)
    
    def grammar_power(self):
        return self.number_of_goals()/self.number_of_goal_strings()
    
    def plotGraph(self):
        GoalGraph.plotGraph(self, text=self.string)


class CompleteGrammarTree(Grammar):
    
    def __init__(self, goal_prob=[0,0,0.5]):
        rules = [('S', 'Sa'), ('S', 'Sb')] 
        goal_strings = []
        for i,p in enumerate(goal_prob):
            for k in range(2**i, 2**(i+1)):
                if random.random() < p:
                    c  = {'0': 'a', '1': 'b'}
                    s  = "S"
                    kb = '{0:b}'.format(k)
                    for i in kb:
                        s = s + c[i]
                    goal_strings.append(s)
        Grammar.__init__(self, 
                         start_string = 'S', 
                         rules        = rules, 
                         goal_strings = goal_strings, 
                         max_depth    = len(goal_prob))



class RandomGrammar(Grammar):

    def __init__(self, 
                 r  = 3,  # number of rules in addition to S->''
                 g  = 3,  # number of goal strings
                 gl = lambda: random.randint(1, 3), #length of goal strings
                 d  = 4):  # maximum depth of tree
        rules = [ ('S', '') ]
        possible_rules = [
            ('S', 'Sa'), 
            ('S', 'Sb'), 
            ('S', 'aS'), 
            ('S', 'bS'),
            ('Sa', 'aS'),
            ('Sb', 'bS'),
            ('aS', 'Sa'),
            ('bS', 'Sb'),
        ]
        rules.extend(random.sample(possible_rules, r))
        goals = [ random_string(['a', 'b'], gl()) for j in range(g) ]
        Grammar.__init__(self, 
                         start_string='S', 
                         goal_strings=goals, 
                         rules=rules, 
                         max_depth=d)
    
    @staticmethod
    def random_init():
        d = random.randint(11, 15)
        g = random.randint(3, 5*d)
        lgl =  random.randint(1, 3)
        ugl = random.randint(lgl, d)
        gl = lambda: random.randint(lgl, ugl)
        r = random.randint(4, 8)
        return dict(
            d  = d,
            g  = g,
            gl = gl,
            r  = r
        )

    feature_list = [
        GoalGraph.degree_mean,
        GoalGraph.degree_stdev,
        GoalGraph.number_of_goals,
        GoalGraph.shallowest_goal,
        GoalGraph.pseudo_diameter,
        GoalGraph.goal_frequency,
        Grammar.number_of_rules,
        Grammar.maximum_depth,
        Grammar.grammar_power,
        #bfs_search,
        #dfs_search,
        GoalGraph.bfs_beats_dfs,
    ]



class CompleteTree(GoalGraph):
    
    def __init__(self, goal_prob=[0,0,0.15,0.1,0], branching_factor=2, create_graph=True):
        GoalGraph.__init__(self)
        self.D = len(goal_prob)-1 
        self.goal_prob_vec = goal_prob

        def add_children(v, d, max_d):
            if goal_prob[d] > random.random(): self.make_goal(v)
            if d == max_d: return # add no more children
            c_list = self.add_vertex(branching_factor) # add two chilren
            for c in c_list:
                self.add_edge(v,c)
                add_children(c, d+1, max_d)
        
        if create_graph: add_children(self.root, 0, self.D)
    
    def depth(self):
        return self.D
    
    def expected_bfs(self):
        p = self.goal_prob_vec
        # level has goal
        l = [ 1-(1-pk)**(2**k) for (k,pk) in enumerate(p) ] 
        #print(l)
        # level is first goal
        m = [ prod([ 1-li for li in l[:k] ])*lk for (k,lk) in enumerate(l) ]
        #print(m)
        # expected search time for finding goal on level k
        lexp = [(k, (2**k - 1 + expectation_truncated_geometric(pk, 2**k)))
                   for (k,pk) in enumerate(p) if pk != 0]
        # if len(lexp) == 0: return 2**D # or raise exception?
        #print(lexp)
        #est = sum([(2**k-1+max(1/pk, 2**k))*m[k] for (k,pk) in enumerate(p) if pk != 0 ]) 
        return sum([lexpk*m[k] for (k,lexpk) in lexp]) or 2**self.D
    
    def old_expected_bfs(self):
        p = self.goal_prob_vec
        # level has goal
        l = [ 1-(1-pk)**(2**k) for (k,pk) in enumerate(p) ] 
        #print(l)
        # level is first goal
        m = [ prod([ 1-li for li in l[:k] ])*lk for (k,lk) in enumerate(l) ]
        #print(m)
        # expected search time 
        est = sum([(2**k-1+min(1/pk, 2**k))*m[k] for (k,pk) in enumerate(p) if pk != 0 ]) 
        return est

    def expected_dfs(self):
        p = self.goal_prob_vec
        if max(p) == 1: return np.argmax(p)+1
        D = self.D
        # level rates
        l = [ -math.log(1-pk) for pk in p ]
        # total rate
        tr = sum([ lk/(2**(D-k+1)) for (k,lk) in enumerate(l) ])
        return 1/max(tr, 2**(-D))

    def predicted_difference(self):
        """DFS-BFS. Negative means DFS better"""
        return self.expected_dfs()-self.expected_bfs()
    
    def predicted_log_ratio(self):
        """log(DFS/BFS). Negative means DFS better"""
        return np.log(self.expected_dfs()/self.expected_bfs())
    
    def prediction_correct(self):
        return ((self.expected_dfs() - self.expected_bfs())*\
               (dfs_search(self)    - bfs_search(self)) > 0)*1.0
    
    feature_list = [
        depth,
        GoalGraph.shallowest_goal,
        expected_bfs,
        #expected_dfs,
        bfs_search
    ]
    
    @staticmethod
    def random_init():
        depth = random.randint(5,7)
        peak = depth-random.randint(1,4)
        interval = random.randint(1,4)
        goal_prob = [ (peak-interval < i < peak+interval)*0.03/depth 
                      for i in range(depth+1)]
        return {
            depth     : depth,
            goal_prob : goal_prob
        }


class CompleteTreeGoalPeak(CompleteTree):
    def __init__(self, depth=5, peak=3, spread=1, max_prob=0.01, create_graph=True):
        self.K = peak
        self.max_p = max_prob
        self.s = spread
        goal_prob = [ max_prob*math.exp( -(peak-k)**2/spread ) 
                      for k in range(depth+1) ]
        CompleteTree.__init__(self, goal_prob=goal_prob, create_graph=create_graph)
    
    def goal_peak(self):
        return self.K

    def goal_spread(self):
        return self.s
    
    def max_prob(self):
        return self.max_p
    
    @staticmethod
    def random_init():
        depth  = random.randint(7,15)
        middle = math.floor(depth/2)
        peak   = random.randint(middle-3, middle+3)
        return {
            depth    : depth,
            peak     : peak,
            spread   : 1,
            max_prob : 0.05
        }
    
    feature_list = [
        CompleteTree.depth,
        #goal_peak,
        #goal_spread,
        #max_prob,
        CompleteTree.expected_bfs,
        bfs_search,
        CompleteTree.expected_dfs,
        dfs_search,
        CompleteTree.prediction_correct
    ]



class CompleteTreeSGL(CompleteTree):
    
    def __init__(self, depth=5, goal_depth=3, goal_prob=0.15, create_graph=True):
        goal_prob_vec = [ 0 for i in range(depth+1)]
        goal_prob_vec[math.ceil(goal_depth)] = goal_prob
        CompleteTree.__init__(self, goal_prob=goal_prob_vec, create_graph=create_graph)
        self.K = goal_depth
        self.p = goal_prob
    
    def goal_depth(self):
        return self.K
        
    def goal_depth_quote(self):
        return self.K/self.D
    
    def goal_prob(self):
        return self.p
                
    def bfs_better(self): # as predicted by theory
        gamma = 1 + np.log2((1-self.p)/self.p)
        K = self.K
        D = self.D
        return (K < D/2 + gamma/2)
    
    def marcus_exp_dfs(self):
        return (1/self.p - 1)*2**(self.D-self.K+1)
    
    feature_list = [
        #number_of_goals,
        CompleteTree.depth,
        CompleteTree.expected_dfs,
        dfs_search,
        #marcus_exp_dfs,
        #goal_depth,
        #goal_prob,
        CompleteTree.expected_bfs,
        bfs_search,
        #lambda g: bfs_search(g)-dfs_search(g),
        #bfs_better
    ]
    
    @staticmethod
    def random_init():
        depth = random.randint(10, 15)
        #goal_depth = random.randint(3,5)
        #goal_prob = random.uniform(0.01, 0.5)
        #depth = 16
        #goal_depth = 12
        goal_depth=depth-3
        goal_prob = 0.07
        return {
            depth      : depth,
            goal_depth : goal_depth,
            goal_prob  : goal_prob
        }


class RandomGraph(GoalGraph):
    """Create a goal graph with randomly sampled nodes at 
       specified distance from root"""

    def __init__(self, 
                 N=40, 
                 goal_level=3, 
                 number_of_goals=2, 
                 degree=lambda: sample_k(5)):
        GoalGraph.__init__(self)
        self.K = goal_level
        self.number_of_goals = number_of_goals
        # Copy a random graph to self
        rgraph = generation.random_graph(N, degree, directed=False)
        self.add_vertex(N)
        for e in rgraph.edges():
            self.add_edge(e.source().__int__(), e.target().__int__())
        # Put goals at the right level (rejection sampling, inefficient)
        # could be made more efficient by BFS layering the graph
        g = 0
        i = 0
        while g < self.number_of_goals and i < 10*N:
            i += 1
            v = self.vertex(random.randint(1, N))
            if topology.shortest_distance(self, self.root, v)==goal_level and \
               not self.goal[v]:
                self.goal[v] = True
                self.goals.append(v)
                g += 1        
        if i >= 10*N:
            print("goaling failed")
    
    @staticmethod
    def random_init():
        N=random.randint(10,100)
        goal_level=random.randint(3,max(3,int(N/10)))
        number_of_goals=random.randint(1,goal_level)
        degree=lambda: sample_k(5)
        return {
            N               : N, 
            goal_level      : goal_level, 
            number_of_goals : number_of_goals, 
            degree          : degree
        }

    # feature_list = [
    #     number_of_goals,
    #     pseudo_diameter,
    #     Goalshallowest_goal,
    #     degree_mean,
    #     degree_stdev,
    #     bfs_search,
    #     dfs_search,
    #     lambda g: bfs_search(g)-dfs_search(g)
    # ]


# Experiment methods

def add_pd_row(df, row):
    df.loc[df.shape()[0]] = row

def getcsv(graph_class, features = None):
    filename = graph_class.__name__+".csv"
    features = features or graph_class.feature_list
    #print("features",features)
    csvfile = csv.writer(open(filename, 'w', 1))
    csvfile.writerow([f.__name__ for f in features])
    return csvfile

def experiment(graph_class, features=None, 
               random_init=None, 
               csvfile = None, # csv.writer(open(filename, 'a', 1))
               append = False,
               iterate=10):
    features = features or graph_class.feature_list
    random_init = random_init or graph_class.random_init
    if append and not csvfile:
        csvfile = csv.writer(open(graph_class.__name__+".csv", 'a', 1))
    else: 
        csvfile = csvfile or getcsv(graph_class, features)
    rows_list = []
    for i in range(iterate):
        g = graph_class(**random_init())
        try: 
            feats = [f(g) for f in features]
            rows_list.append(feats)
            csvfile.writerow(feats)
            print(i, end=" ", flush=True)
            #g.plotGraph()
        except NameError as err:
            #print("exception", err)
            print('X', end=" ", flush=True)
    df = pandas.DataFrame(rows_list, columns=[f.__name__ for f in features])
    return df

experiment(RandomGrammar, iterate=2000)


def averages(df):
    return pandas.DataFrame([[round(np.average(df[c]), 2) for c in df]], columns=df.columns.values)

def correlations(df):
    return np.round(np.corrcoef([df[c] for c in df]), 2)

# TODO: Move to predict or delete?
def my_contour(diff, name=""):
    plt.figure()
    plt.gca().patch.set_color('.25')
    cs = plt.contourf(diff, (-50, -3, 3, 50), cmap='Greys') #better levels?
    plt.contour(diff, (-1,0,1))
    #cs = plt.contourf(diff, )
    plt.colorbar(cs)
    plt.xlabel('Depth')
    plt.ylabel('Goal peak')
    plt.legend('')
    plt.title('Decision boundary '+name+': Negative=DFS better')
    #proxy = [plt.Rectangle((0,0),1,1,fc = pc.get_facecolor()[0]) 
            #for pc in cs.collections]
    #plt.legend(proxy, ["range()", "range(3-4)", "range(4-6)"])

# TODO: Generalize with init parameters (and two variables to vary?)
def graph_experiment(graph_class, D, kvargs={}):
    """ Method for analytic experiments """
    diff = np.empty((D, D))
    diff[:] = np.NAN
    for i in range(0, D):
        d = i+5
        for k in range(D):
            g = graph_class(d, d*k/D, create_graph=False, **kvargs)
            diff[k][i] = g.predicted_log_ratio()
    return diff
    my_contour(diff, name=graph_class.__name__)


# Conclusions:
# - theory predicts well when the goal is sufficiently deep
#   and improbable (depth > 10, prob < 0.15).
# - when either the prob is high > 0.5, or the goal is shallow (3-5),
#   the theory predicts poorly (expected value does not correlate
#   with actual outcome).






# def complete_tree_exp(n=10):
#     feature_list = [
#         number_of_goals,
#         #pseudo_diameter,
#         #shallowest_goal,
#         #degree_mean,
#         #degree_stdev,
#         CompleteTree.depth,
#         CompleteTree.goal_depth,
#         CompleteTree.goal_prob,
#         bfs_search,
#         dfs_search,
#         lambda g: bfs_search(g)-dfs_search(g),
#         CompleteTree.bfs_better
#     ]
#     csvfile = getcsv('complete_tree.csv', feature_list)
#     for i in range(n):
#         depth = random.randint(5, 15)
#         goal_depth = random.randint(3,depth)
#         #goal_prob = random.uniform(0.01, 0.5)
#         goal_prob = 0.07
#         g = CompleteTree(depth, goal_depth, goal_prob)
#         try: 
#             feats = g.features(feature_list)
#             csvfile.writerow(feats)
#             print(i, feats)
#         except NameError as err:
#             print(err)




        

# complete_tree_exp(1000)


# Run experiment
# for i in range(100):
#     g = MyGraph()
#     features = g.features()
#     bfs = BFSSearch(g)
#     dfs = DFSSearch(g)
#     features.extend([bfs, dfs])
#     features.insert(0,i)
#     print(features)


#g = random_graph(N, lambda: sample_k(5), directed=False)

#g = random_graph(N, lambda: sample_k(5), model="probabilistic", vertex_corr=lambda i, k: 1.0 / (1 + abs(i - k)), directed=False,n_iter=100)

#print(g.list_properties())

#for v in g.vertices():
#   print(v, goal[v])

#pos = sfdp_layout(g)
#pos = fruchterman_reingold_layout(g)
#pos = arf_layout(g)
#pos = radial_tree_layout(g,g.vertex(0))
#pos = random_layout(g)
#pos = arf_layout(g, pos=pos)
#pos = fruchterman_reingold_layout(g, pos=pos)

# root = g.new_vertex_property("bool")
# root[g.vertex(0)] = True

#g = MyGraph()

#dist_hist = graph_tool.stats.distance_histogram(g)
#deg_hist = graph_tool.stats.vertex_hist(g,"total")




# def sample_k(max):
#     accept = False
#     while not accept:
#         k = random.randint(1,max+1)
#         accept = random.random() < 1.0/k
#     return k


# class MyGraph(gt.Graph):
#     def __init__(self, N=50, number_of_goals=3):
#         gt.Graph.__init__(self)
#         self.N = N
#         self.number_of_goals = number_of_goals
#         self.g = generation.random_graph(N, lambda: sample_k(5), directed=False)
#         self.root = self.g.vertex(0)
#         self.goal = self.g.new_vertex_property("bool")
#         def put_goals():
#             self.goals = random.sample(range(N), number_of_goals)
#             for i in self.goals:
#                 self.goal[self.g.vertex(i)] = True
#             if self.shallowest_goal() > N+1:
#                 put_goals()
#         put_goals()

#     def pseudo_diameter(self):
#         return topology.pseudo_diameter(self.g)[0]

#     def shallowest_goal(self):
#         m = min([topology.shortest_distance(self.g, self.root, 
#                                             self.g.vertex(v), 
#                                             max_dist=self.N+1) 
#                  for v in self.goals])
#         if m == self.N+1:
#             return 0
#         else:
#             return m

#     def degree_histogram(self):
#         return stats.vertex_hist(self.g, "total")

#     def features(self):
#         return [
#             self.shallowest_goal(),
#             self.pseudo_diameter(),
#             hist2mean(self.degree_histogram()),
#             hist2stdev(self.degree_histogram())
#         ]
#         # return {
#         #     'diam': self.pseudo_diameter(),
#         #     'deg_mean': hist2mean(self.degree_histogram()),
#         #     'deg_stdev': hist2stdev(self.degree_histogram())
#         # }

#     def plotGraph(self):
#         g = self.g
#         pos = gt.arf_layout(g)
#         pos[g.vertex(0)] = [0.1,0.1]
#         pos = gt.arf_layout(g, pos=pos)
#         color = g.new_vertex_property("string")
#         for v in g.vertices():
#             if self.goal[v]:
#                 color[v] = 'g'
#             else:
#                 color[v] = 'w'
#             color[g.vertex(0)] = 'b' # root
#         gt.graph_draw(g, pos, vertex_color=color, 
#                    vertex_fill_color=color, vertex_text=g.vertex_index, 
#                    vertex_font_size=18, output_size=(800, 800))
    
#     def plotFeatures(self):
#         plt.hist(self.degree_histogram())
#         plt.show()
        
#### End of MyGraph #####
