import math
import numpy as np
import matplotlib.pyplot as plt
from itertools import combinations as comb
from argparse import ArgumentParser
from scipy.stats.stats import pearsonr
import pandas
from sklearn import svm, feature_selection
from sklearn.metrics import mean_squared_error
import timeit

kernels = [
    svm.SVC(kernel='linear', C=1, cache_size=12000),
    svm.SVC(kernel='linear', C=1e4, cache_size=12000),
    svm.SVC(kernel='poly', degree=2, coef0=1, C=1, cache_size=12000),
    svm.SVC(kernel='poly', degree=2, coef0=1, C=1e4, cache_size=12000),
    svm.SVC(kernel='poly', degree=3, coef0=1, C=1e4, cache_size=12000),
    svm.SVC(kernel='poly', degree=2, coef0=1, C=1e4, cache_size=12000),
    svm.SVC(kernel='rbf', gamma=1, C=10**4, cache_size=12000),
    svm.SVC(kernel='rbf', gamma=0.001, C=10**4, cache_size=12000),
    svm.SVC(kernel='rbf', gamma=1000, C=10**4, cache_size=12000),
    svm.SVC(kernel='sigmoid', gamma=1, coef0=1, C=10**4, cache_size=12000),
    svm.SVC(kernel='sigmoid', gamma=0.001, coef0=1, C=10**4, cache_size=12000),
    svm.SVC(kernel='sigmoid', gamma=1000, coef0=1, C=10**4, cache_size=12000),
]


rbfkernels = [svm.SVC(kernel='rbf', gamma=g, C=c, cache_size=12000)
              for g in [1e-4, 1e-3, 1e-2, 1e-1, 1e0, 1e1, 1e2, 1e3]
              for c in [1e-4, 1e-3, 1e-2, 1e-1, 1e0, 1e1, 1e2, 1e3]]


def file_to_dataframe(filename, normalize=False):
    """Read the given file into a pandas dataframe"""
    data = pandas.io.parsers.read_csv(filename)
    if normalize:
        return normalize_data(data)
    else:
        return data


def normalize_data(df):
    """Normalize all columns to mean 0, stdev 1"""
    ndf = pandas.DataFrame()
    for cname in df:
        c = df[cname]
        mean = np.mean(c)
        std = np.std(c)
        nc = (c-mean)/std
        ndf[cname] = nc
    return ndf


def learning_columns(df, cols=None):
    """Extract the learning columns from pandas dataframe"""
    cols = cols or list(range(df.shape[1]-1))
    return df[[df.columns[i] for i in cols]]


def target_column(df, col=None, regression=False):
    """Extract a target column from a pandas dataframe as pandas Series.
    The column is binarized unless regression is true."""
    col = col or len(df.columns)-1
    return df[df.columns[col]] if regression else df[df.columns[col]] > 0


def split_data(df, lcols=None, tcol=None, regression=False):
    """Split data into learning and testing data,
    and learning and target columns"""
    lcols = lcols or list(range(df.shape[1]-1))
    tcol = tcol or df.shape[1] - 1
    middle = df.shape[0]//2
    l = df[:middle]
    t = df[middle:]
    ldata = learning_columns(l, lcols)
    tdata = learning_columns(t, lcols)
    ltarg = target_column(l, tcol, regression=regression)
    ttarg = target_column(t, tcol, regression=regression)
    return (ldata, ltarg, tdata, ttarg)


def test_prediction(clf, data, target):
    """Test classifier's performance on data->target """
    correct = (clf.predict(data) == target)
    return sum(correct)/len(correct)


def test_correlation(clf, data, target):
    """Test regressor's performance on data->target by correlation coef"""
    return pearsonr(clf.predict(data), target)[0]


def test_rmse(clf, data, target):
    """Test regressor's performance on data->target by RMSE"""
    return math.sqrt(mean_squared_error(clf.predict(data), target))


def constant_prediction(testtarget):
    """Give a baseline by how well a classifier with no data performs"""
    s = sum(testtarget)/len(testtarget)
    return max(s, 1-s)


def constant_rmse(target):
    """ """
    return math.sqrt(mean_squared_error([np.mean(target)]*len(target), target))


def plot_classifier(clf, data, target, features=None, xdim=None, ydim=None):
    """
    Plot a decision-boundary plot of the two given features with rest of
    values fixed to their average.
    """
    features = features or data.columns.tolist()
    assert(len(features) == 2)
    dx = data[features[0]]
    dy = data[features[1]]
    xmax, xmin = xdim or (dx.min(), dx.max())
    ymax, ymin = ydim or (dy.min(), dy.max())
    x = np.linspace(xmin, xmax)
    y = np.linspace(ymin, ymax)
    xx, yy = np.meshgrid(x, y)
    m = np.zeros((len(xx.ravel()), len(data.columns)))
    for i, c in enumerate(data. columns):
        if c == features[0]:
            m[:, i] = xx.ravel()
        elif c == features[1]:
            m[:, i] = yy.ravel()
        else:
            m[:, i] = np.mean(data[c])
    z = clf.predict(m).reshape(xx.shape)
    plt.contourf(x, y, z)
    plt.scatter(dx, dy, c=target, cmap=plt.cm.Paired)
    plt.xlim(x.min(), x.max())
    plt.ylim(y.min(), y.max())
    plt.xlabel(features[0])
    plt.ylabel(features[1])
    plt.show()


class Learner():
    """
    Learner object, initialized with a file and optional learning and
    target columns
    """

    def __init__(self,
                 filename  = 'data/random_grammar.csv',
                 lcols     = None,  # e.g. [0,1,2,3,4]
                 tcol      = None,  # e.g. 5
                 clf_gen   = None,  # e.g. lambda: svm.SVC(kernel='linear')
                 normalize = False,
                 regression= False,
    ):
        """Initialize Learner with a csv data file, and optional
        learning columns, target column, and classifier generator"""
        self.regression = regression,
        self.normalize = normalize,
        # Read data
        self.data = file_to_dataframe(filename, normalize=normalize)
        self.l = learning_columns(self.data, cols=lcols)
        self.t = target_column(self.data, col=tcol)
        (self.ll, self.lt, self.tl, self.tt) \
            = split_data(self.data, lcols=lcols, tcol=tcol,
                         regression=regression)
        self.lcols = self.ll.columns.tolist()
        if regression:
            self.cons_perf = constant_rmse(self.tt)
            print(self.cons_perf)
            self.perf = test_rmse
            self.clf_gen = clf_gen or (lambda: svm.SVR(kernel='linear'))
        else:
            self.cons_perf = constant_prediction(self.tt)
            self.perf = test_prediction
            self.clf_gen = clf_gen or (lambda: svm.SVC(kernel='linear'))
        print("Learning from " + str(self.lcols) +
              "\nwith baseline %.2f " % self.cons_perf)

    def do_all(self, ass=True):
        """Run all classifications, prints performance"""
        print("training full classifier")
        self.train_full_clf()
        print("full classifier done, perf: %.2f" % self.full_perf)
        self.feature_rank()
        print("\nfeatures ranked:\n" + str(self.feat_ranked))
        self.train_on_best_pair()
        print("\nbest pair: " + str(self.feat_pairs[0]) +
              " gives %.2f " % self.red_perf)
        self.train_on_optimal()
        print("\noptimal features:\n" + str(self.opt_feats))
        removed = set(self.lcols).difference(self.opt_feats)
        print("removed features: " + str(removed))
        # if ass: assert self.full_perf > self.red_perf
        # if ass: assert self.full_perf > self.cons_perf
        print('\nPerformance ' + self.perf.__name__)
        print('Baseline:                    %.2f' % self.cons_perf)
        print('Full classifier performance: %.2f' % self.full_perf)
        print('Best two performance:        %.2f' % self.red_perf)
        print('Best four performance:       %.2f' % self.evaluate_tuple(self.feat_ranked[:4]))
        print('Optimal feats performance:   %.2f' % self.opt_perf)

    def do_full_only(self, clf=None):
        clf = clf or self.clf_gen()
        self.train_full_clf(clf)
        print('\nPerformance ' + self.perf.__name__)
        print('Baseline:                    %.2f' % self.cons_perf)
        print('Full classifier performance: %.2f' % self.full_perf)
        print('with clf:\n', clf)

    def do_all_kernels(self, kernels):
        for k in rbfkernels:
            t = timeit.timeit(lambda: self.do_full_only(clf=k), number=1)
            print("Time: ", t, "\n")

    def train_full_clf(self, clf=None):
        """Train a classifier on all learning columns"""
        self.fclf = clf or self.clf_gen()
        self.fclf.fit(self.ll, self.lt)
        self.full_perf = self.perf(self.fclf, self.tl, self.tt)

    def train_on_tuple(self, tup, clf=None):
        """Train a classifier on the given tuple of columns"""
        clf = clf or self.clf_gen()
        tup = list(tup) if isinstance(tup, tuple) else tup
        clf.fit(self.ll[tup], self.lt)
        return clf

    def evaluate_tuple(self, tup, clf=None):
        """Evaluate how well the given tuple can predict target"""
        clf = self.train_on_tuple(tup, clf=clf)
        return self.perf(clf, self.tl[list(tup)], self.tt)

    def train_on_best_pair(self, clf=None):
        """Train classifier on the best two features, as determined by
        feature ranking."""
        self.rclf = clf or self.clf_gen()
        self.feature_rank() if not hasattr(self, "feat_pairs") else ""
        self.rll = self.ll[list(self.feat_pairs[0])]
        self.rtl = self.tl[list(self.feat_pairs[0])]
        self.rclf.fit(self.rll, self.lt)
        self.red_perf = self.perf(self.rclf, self.rtl, self.tt)
        return self.rclf

    def train_on_optimal(self, clf=None):
        """Train only on optimal features, as determined by RFECV."""
        self.oclf = clf or self.clf_gen()
        self.optimal_features() if not hasattr(self, "opt_feats") else ""
        self.oll = self.ll[self.opt_feats]
        self.otl = self.tl[self.opt_feats]
        self.oclf.fit(self.oll, self.lt)
        self.opt_perf = self.perf(self.oclf, self.otl, self.tt)
        return self.oclf

    def feature_rank(self, clf=svm.SVC(kernel="linear")):
        """Use the built-in Recursive Feature Selection model"""
        rfe = feature_selection.RFE(estimator=clf,
                                    n_features_to_select=1, step=1)
        rfe.fit(self.ll, self.lt)
        sorted_cols = [x for (y, x) in
                       sorted(zip(rfe.ranking_, self.ll.columns))]

        def priority(t):
            """Ranks feature pairs by the sum of their rankings"""
            i = self.lcols.index(t[0])
            j = self.lcols.index(t[1])
            return rfe.ranking_[i] + rfe.ranking_[j]

        col_pairs = [(x, y) for i, x in enumerate(self.lcols)
                     for y in self.lcols[i+1:]]
        self.feat_pairs = sorted(col_pairs, key=priority)
        self.feat_ranked = sorted_cols
        return self.feat_pairs

    def optimal_features(self, clf=svm.SVR(kernel='linear')):
        """Recursively decide which features
        improve the fit by cross-validation"""
        rfecv = feature_selection.RFECV(estimator=clf)
        rfecv.fit(self.l, self.t)
        best_cols = [x for x, y in zip(self.lcols, rfecv.ranking) if y == 1]
        self.opt_feats = best_cols
        return self.opt_feats

    def plot_n_best_feat_pairs(self, n=5):
        """ Plot the n best features with rest of data at mean values"""
        self.feature_rank() if not hasattr(self, "feat_pairs") else ""
        for t in self.feat_pairs[:n]:
            plot_classifier(self.fclf, self.tl, self.tt, t)
            print(str(t)+" achieves %.2f" % self.evaluate_tuple(t))


# If run as script:
if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("filename")
    parser.add_argument("-l", "--lcols")
    parser.add_argument("-t", "--target", type=int)
    parser.add_argument("-c", "--classifier")
    parser.add_argument("-r", "--regression", action="store_true")
    parser.add_argument("-a", "--all_kernels", action="store_true")
    args = parser.parse_args()
    filename = args.filename
    lcols = [int(i) for i in args.lcols.split()] if args.lcols else None
    clf = (lambda: eval(args.classifier)) if args.classifier else None
    l = Learner(filename=args.filename, lcols=lcols,
                tcol=args.target, clf_gen=clf, regression=args.regression)
    if args.all_kernels:
        l.do_all_kernels(kernels)
        # TODO: Add for regression mode
    else:
        l.do_full_only()


####################################################
#
# Deprecated methods
#
####################################################

# def svmfit(data, target, clf = default_clf()):
#     #clf = svm.SVC(kernel='poly', degree=2)
#     clf.fit(data, target)
#     #print('predicting from ', list(data.columns))
#     return clf

# # TODO: What does C do?
# def feature_select(learn, target, clf=svm.SVC(kernel="linear")):
#     """
#     Use the built-in Recursive Feature Selection model.
#     Appears to only work with linear kernel.
#     """
#     rfe = feature_selection.RFE(estimator=clf,
#                                 n_features_to_select=1, step=1)
#     rfe.fit(learn, target)
#     return [c for i,c in enumerate(learn.columns) if rfe.ranking_[i]==1]
#     #return rfe

# def feature_select2(learn, target, clf = default_clf()):
#     """
#     Select the two best features by testing which two offers the best
#     learning rate.
#     Could be improved with separate test set.
#     """
#     col_pairs = list(comb(learn.columns, 2))
#     perf = {}
#     for (c1, c2) in col_pairs:
#         clf.fit(learn[[c1,c2]], target)
#         perf[(c1,c2)] = self.perf(clf, learn[[c1, c2]], target)
#     best = col_pairs[0]
#     for k in perf:
#         if perf[k]>perf[best]:
#             best = k
#         print(k,perf[k])
#     return [best[0], best[1]]


    # for use in Learner. RFE based seems to work better
    # def feature_rank2(self, tuple_size=2, clf=svm.SVC(kernel='linear')):
    #     """
    #     Rank feature pairs by learning rate.
    #     Could be improved with separate test set.
    #     """
    #     col_pairs = list(comb(self.ll.columns, tuple_size))
    #     self.perf = {}
    #     for tup in col_pairs:
    #         #print("ranking", tup)
    #         clf.fit(self.ll[list(tup)], self.lt)
    #         self.perf[tup] \
    #             = self.perf(clf, self.ll[list(tup)], self.lt)
    #     self.feat_pairs = sorted(self.perf, key=self.perf.get)
    #     return self.feat_pairs


# we create 40 separable points
#np.random.seed(0)
#X = np.r_[np.random.randn(20, 2) - [2, 2], np.random.randn(20, 2) + [2, 2]]
#Y = [0] * 20 + [1] * 20

# data = pandas.io.parsers.read_csv('complete_tree.csv')
# X = data.as_matrix(columns=['depth', 'goal_depth'])
# Y = np.squeeze(data.as_matrix(columns=['<lambda>']))>0

# # fit the model
# clf = svm.SVC(kernel='linear')
# clf.fit(X, Y)

# def default_clf():
#     return svm.SVC(kernel='linear')
#     #return svm.SVC(kernel='poly', degree=2)

# # TODO: Change to plot_linear_classifier ?
# def plot(clf):
#     # get the separating hyperplane
#     w = clf.coef_[0]  # normal vector
#     a = -w[0] / w[1]
#     b = -(clf.intercept_[0]) / w[1]
#     xx = np.linspace(X.min(), X.max())
#     yy = a * xx + b
#     gp = 0.07
#     gamma = 1 + np.log2((1-gp)/gp)
#     yt = 0.5 * xx + gamma/2

#     print("separator coeffecients:", clf.coef_[0])
#     print("separator intercept:", clf.intercept_[0])
#     print("separating line: y = "+str(a)+" x + "+str(b))

#     # plot the parallels to the separating hyperplane that pass through the
#     # support vectors
#     #b = clf.support_vectors_[0]
#     #yy_down = a * xx + (b[1] - a * b[0])
#     #b = clf.support_vectors_[-1]
#     #yy_up = a * xx + (b[1] - a * b[0])

#     # plot the line, the points, and the nearest vectors to the plane
#     plt.plot(xx, yy, 'k-')
#     plt.plot(xx, yt, 'r--')
#     #plt.plot(xx, yy_down, 'k--')
#     #plt.plot(xx, yy_up, 'k--')

#     plt.scatter(clf.support_vectors_[:, 0], clf.support_vectors_[:, 1],
#                 s=80, facecolors='none')
#     plt.scatter(X[:, 0], X[:, 1], c=Y, cmap=plt.cm.Paired)
#     plt.xlabel('Tree depth')
#     plt.ylabel('Goal depth')
#     plt.legend(['SVM separator y = %.2f x + %.2f' % (a, b),
#                 'Theory tree_and_grammar y = 0.5x + %.2f' % (gamma/2),
#                 'DFS better', 'BFS Better'], loc='upper left')
#     plt.title('DFS vs BFS in complete tree, goal_prob='+str(gp)+', #data='+str(data.shape[0]))

#     plt.axis('tight')
#     plt.gray()
#     plt.show()
