#
# Author:  Tom Everitt
# Email:   tom4everitt@gmail.com
# website: http://tomeveritt.se
#
# License:
# Use as you like as long as you credit me in publications.
#

import csv
from argparse import ArgumentParser

import numpy as np
import pandas

import safe_experiments
from graphs.binary_grammar import BinaryGrammar
from graphs.complete_tree import CompleteTree
from graphs.complete_tree_sgl import CompleteTreeSGL
from search_methods import bfs_search, dfs_search


def add_pd_row(df, row):
    df.loc[df.shape()[0]] = row


def getcsv(graph_class, features=None, outfile=None):
    filename = outfile or graph_class.__name__ + ".csv"
    features = features or graph_class.feature_list
    #print("features",features)
    csvfile = csv.writer(sf.open(filename), delimiter='\t')
    csvfile.writerow([f.__name__ for f in features])
    return csvfile


def experiment(graph_class, features=None,
               random_init=None,
               csvfile=None,  # csv.writer(open(filename, 'a', 1))
               outfile=None,
               append=False,
               iterate=10):
    """Run experiment with given graph class, and output piped to
    given file"""
    outfile = outfile or graph_class.__name__ + ".csv"
    features = features or graph_class.feature_list
    random_init = random_init or graph_class.random_init
    if append and not csvfile:
        csvfile = csv.writer(sf.open(outfile, mode='a'), delimiter='\t')
    else:
        try:
            csvfile = csvfile or getcsv(graph_class,
                                        features, outfile=outfile)
        except FileExistsError as err:
            print("file exists. Remove first or use -a to append to it.")
            return
    rows_list = np.zeros((0, len(features)))
    bfs_win = 0
    print(iterate)
    # TODO: Multi-thread?
    for i in range(iterate):
        g = graph_class(**random_init())
        try:
            feats = [f(g) for f in features]
            rows_list = np.append(rows_list, np.array([feats]), axis=0)
            csvfile.writerow(feats)
            print(i, end=" ", flush=True)
        except NameError as err:
            print('X', end=" ", flush=True)
    df = pandas.DataFrame(rows_list, columns=[f.__name__ for f in features])
    return df


def averages(df):
    return pandas.DataFrame([[round(np.average(df[c]), 2)
                              for c in df]], columns=df.columns.values)


def correlations(df):
    return np.round(np.corrcoef([df[c] for c in df]), 2)


def perc_err(correct, guess):
    """The error of the guess in percentage"""
    return 100*np.abs(correct-guess)/correct

##########################################################
#
# Methods to verify analytical approximations.
# (Uncomment calls to run.)
#
##########################################################


def verifySGL(krange, prange, D=15, iterate=1000):
    feature_list = [
        CompleteTreeSGL.sgl_expected_dfs,
        dfs_search,
        CompleteTreeSGL.sgl_expected_bfs,
        bfs_search,
    ]
    dfsana = np.zeros((len(krange), len(prange)))
    dfsexp = np.zeros((len(krange), len(prange)))
    dfserr = np.zeros((len(krange), len(prange)))

    bfsana = np.zeros((len(krange), len(prange)))
    bfsexp = np.zeros((len(krange), len(prange)))
    bfserr = np.zeros((len(krange), len(prange)))
    for (ki, k) in enumerate(krange):
        print('k', k)
        for (pi, p) in enumerate(prange):
            #if p < 2**(-k-2): continue
            print('k p', k, p)
            random_init = lambda: dict(depth=D, goal_depth=k, goal_prob=p)
            df = experiment(CompleteTreeSGL,
                            features=feature_list,
                            outfile = 'SGL_D'+str(D)+'p'+str(p)+'k'+str(k),
                            random_init=random_init, iterate=iterate, append=True)
            av = averages(df)
            dfsana[ki,pi] = av['sgl_expected_dfs']
            dfsexp[ki,pi] = av['dfs_search']
            dfserr[ki,pi] = perc_err(av['dfs_search'], av['sgl_expected_dfs'])
            bfsana[ki,pi] = av['sgl_expected_bfs']
            bfsexp[ki,pi] = av['bfs_search']
            bfserr[ki,pi] = perc_err(av['bfs_search'], av['sgl_expected_bfs'])

            print('av\n', av)
    return ( [bfsexp, bfsana, bfserr], [dfsexp, dfsana, dfserr] )

# Verify SGL
# sf = safe_experiments.SafeFiles('sgl')
# D = 14
# krange = [5, 8, 11, 14]
# prange = [0.001, 0.003, 0.01, 0.03, 0.1, 0.3]
# verifySGL(krange, prange, D=D, iterate=10)


def verifyMGL(krange, srange, D=15, iterate=1000):
    feature_list = [
        CompleteTree.expected_dfs,
        dfs_search,
        CompleteTree.expected_bfs_g,
        bfs_search,
    ]
    dfsana = np.zeros((len(krange), len(srange)))
    dfsexp = np.zeros((len(krange), len(srange)))
    dfserr = np.zeros((len(krange), len(srange)))

    bfsana = np.zeros((len(krange), len(srange)))
    bfsexp = np.zeros((len(krange), len(srange)))
    bfserr = np.zeros((len(krange), len(srange)))
    for (ki, k) in enumerate(krange):
        print('k', k)
        for (si, s) in enumerate(srange):
            #if p < 2**(-k-2): continue
            print('k s', k, s)
            random_init = lambda: dict(depth=D, peak=k, spread=s)
            df = experiment(CompleteTreeMGL,
                            features=feature_list,
                            outfile = 'MGL_D'+str(D)+'k'+str(k)+'s'+str(s),
                            random_init=random_init,
                            iterate=iterate, append=True)
            av = averages(df)
            dfsana[ki,si] = av['expected_dfs']
            dfsexp[ki,si] = av['dfs_search']
            dfserr[ki,si] = perc_err(av['dfs_search'], av['expected_dfs'])
            bfsana[ki,si] = av['expected_bfs_g']
            bfsexp[ki,si] = av['bfs_search']
            bfserr[ki,si] = perc_err(av['bfs_search'], av['expected_bfs_g'])

            print('av\n', av)
    return ( [bfsexp, bfsana, bfserr], [dfsexp, dfsana, dfserr] )

# Verify MGL
# sf = safe_experiments.SafeFiles('mgl')
# D = 14
# krange = [5, 8, 11, 14]
# srange = [0.1, 1, 10, 100]
# verifyMGL(krange, srange, D=D, iterate=10)




def verifyCG(krange, prange, D=15, iterate=1000):
    feature_list = [
        BinaryGrammar.expected_dfsl,
        dfs_search,
        BinaryGrammar.expected_dfsu,
        BinaryGrammar.expected_dfs,
        BinaryGrammar.expected_bfs,
        bfs_search,
    ]
    ldfsana = np.zeros((len(krange), len(prange)))
    ldfsexp = np.zeros((len(krange), len(prange)))
    ldfserr = np.zeros((len(krange), len(prange)))

    udfsana = np.zeros((len(krange), len(prange)))
    udfsexp = np.zeros((len(krange), len(prange)))
    udfserr = np.zeros((len(krange), len(prange)))

    dfsana = np.zeros((len(krange), len(prange)))
    dfsexp = np.zeros((len(krange), len(prange)))
    dfserr = np.zeros((len(krange), len(prange)))

    bfsana = np.zeros((len(krange), len(prange)))
    bfsexp = np.zeros((len(krange), len(prange)))
    bfserr = np.zeros((len(krange), len(prange)))
    for (ki, k) in enumerate(krange):
        print('k', k)
        for (pi, p) in enumerate(prange):
            print('k p', k, p)
            random_init = lambda: dict(depth=D, g=k, pg=p)
            df = experiment(BinaryGrammar,
                            features=feature_list,
                            outfile = 'SGL_D'+str(D)+'p'+str(p)+'k'+str(k),
                            random_init=random_init, iterate=iterate, append=True)
            av = averages(df)
            ldfsana[ki,pi] = av['expected_dfsl']
            ldfsexp[ki,pi] = av['dfs_search']
            ldfserr[ki,pi] = perc_err(av['dfs_search'], av['expected_dfsl'])
            udfsana[ki,pi] = av['expected_dfsu']
            udfsexp[ki,pi] = av['dfs_search']
            udfserr[ki,pi] = perc_err(av['dfs_search'], av['expected_dfsu'])
            dfsana[ki,pi] = av['expected_dfs']
            dfsexp[ki,pi] = av['dfs_search']
            dfserr[ki,pi] = perc_err(av['dfs_search'], av['expected_dfs'])
            bfsana[ki,pi] = av['expected_bfs']
            bfsexp[ki,pi] = av['bfs_search']
            bfserr[ki,pi] = perc_err(av['bfs_search'], av['expected_bfs'])

            print('av\n', av)
    return ( [bfsexp, bfsana, bfserr], [ldfsexp, ldfsana, ldfserr], [udfsexp, udfsana, udfserr], [dfsexp, dfsana, dfserr] )

# Verify CG
# sf = safe_experiments.SafeFiles('cg')
# D = 14
# krange = [5, 8, 11, 14]
# prange = [0.001, 0.01, 0.1]
# verifyCG(krange, prange, D=D, iterate=10)



##################################################
#
# If called as script
# Example call: python3 -i 10 CompleteTreeSGL
# Problems are generated according to the class's random_init method,
# and run with the methods in the class's feature_list
#
##################################################

if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("class_name")
    parser.add_argument("-i", "--iterate", type = int)
    parser.add_argument("-o", "--outfile")
    parser.add_argument("-f", "--features")
    parser.add_argument("-a", "--append",  action="store_true")
    args = parser.parse_args()
    class_name = eval(args.class_name)
    features = eval(args.features or 'None')
    sf = safe_experiments.SafeFiles(args.class_name)
    i = args.iterate or 10
    experiment(class_name, features=features, iterate=i,
               outfile=args.outfile, append=args.append)
