#
# Author:  Tom Everitt
# Email:   tom4everitt@gmail.com
# website: http://tomeveritt.se
#
# License:
# Use as you like as long as you credit me in publications.
#

import numpy as np
from scipy.misc import comb
import matplotlib.pyplot as plt

def exp_trunc_geo(p, m):
    """
    Truncated geometric expectation.
    Let X ~ Geo(p), i.e. P(X=k)=(1-p)^(k-1)*p. Return E[ X | X <= m].
    """
    assert 0 <= p <= 1
    if p < 0.00000001: return m/2 + 1 # approximate to avoid division by zero
    return (1 - (1-p)**m*(p*m+1) )/(p*(1-(1-p)**m))

def prod(vec):
    """The product of all the elements in the vector (analogous to sum)"""
    p = 1
    for el in vec:
        p = p*el
    return p

def nCk(n,k):
    """ n choose k """
    assert n >= k
    assert k >= 0
    return int(np.round(comb(n,k,exact=False)))

def lv(D, L):
    """
    Matrix (D+1)x(D+1), element (n,d) is the number of elements on level d
    reachable from a^n given function L(D,n,d)
    """
    return np.array([ [ L(D,n,d) for d in range(D+1) ] for n in range(D+1) ])

def sv(D, L):
    """Vector where element n is the size of subtree under a^n"""
    return sum(lv(D, L).T)

def tv(D, L):
    """Vector where element n is the number of new nodes seen from a^n"""
    s = sv(D, L)
    return np.append( s[:-1]-s[1:], L(D,D,D) )

def av(D, L):
    """Matrix where element (n,d) is the new nodes seen on level d from a^n"""
    lastrow = [0]*(D+1)
    lastrow[D] = L(D,D,D)
    return np.append(lv(D, L)[:-1]-lv(D, L)[1:], [lastrow], axis=0) 

def lmbdav(D, p, L):
    """Vector of length D+1 where element n is the probability of finding goal under a^n"""
    a = av(D, L)
    return [ 1 - prod([(1-p[d])**a[n,d] for d in range(n,D+1)]) for n in range(D+1) ]

def prob_first_goalv(D, p, L):
    """Element n is the probability that the first goal is under a^n"""
    l = lmbdav(D, p, L)
    p = [ l[n]*prod([ (1-lk) for lk in l[n+1:D+1] ]) for n in range(D+1)]
    return p/sum(p)

def lower_tdfs(D, p, L):
    """Lower bound on expected DFS search time"""
    assert D+1 == len(p)
    f = list(prob_first_goalv(D, p, L))
    f.insert(0, 1-sum(f)) # \phi_{-1}
    f = np.array(f)
    s = list(sv(D, L))
    s.append(0) # S_{D+1}
    s = np.array(s)
    #return sum([ s[n+1]*f[n] for n in range(0,D) ])
    return np.dot(s, f)
    
def upper_tdfs(D, p, L):
    """Upper bound on expected DFS search time"""
    assert D+1 == len(p)
    f = list(prob_first_goalv(D, p, L))
    f.insert(0, 1-sum(f))
    f = np.array(f)
    s = list(sv(D, L))
    s.insert(0, s[0]+1)
    s = np.array(s)
    #return sum([ s[n]*f[n] for n in range(0,D+1) ])
    return np.dot(s, f)

def tbfs(D, p, L):
    """Expected BFS search time"""
    l = lv(D+1, L)
    #if extra_level:
    #    p = [ p[k]/2**k for k in range(len(p)) ]
    p = list(p)
    p.append(1)
    # probability level has a goal
    g = np.array([ 1-(1-pk)**l[0,k] for (k,pk) in enumerate(p) ])
    # probability level has the first goal
    f = np.array([ g[n]*prod( 1- g[:n]) for n in range(D+2)])
    assert sum(f) == 1
    s = [ sum(l[0,:n]) for n in range(D+2) ]
    # expected search time on level n, given goal on level n
    lt = np.array([ s[n] + exp_trunc_geo(p[n], l[0,n]) for n in range(D+2) ])
    return np.dot(f, lt)

def tbfs_extra_level(D, p, L):
    return tbfs(D, p, L, extra_level=True)

def tdfs_u_clust(D, g, p):
    pv = [0]*(D+1)
    pv[g] = p
    return upper_tdfs(D, pv, l_grammar_clustered)

def tdfs_l_clust(D, g, p):
    pv = [0]*(D+1)
    pv[g] = p
    return lower_tdfs(D, pv, l_grammar_clustered)

def tbfs_clust(D, g, p):
    pv = [0]*(D+1)
    pv[g] = p
    return tbfs(D, pv, l_grammar_clustered)


def l_grammar_clustered(D, n, d):
    """Number of clustered nodes at level d reachable from a^n"""
    assert n >= 0
    return sum([ nCk(d, i) for i in range(d-n+1) ])

def l_grammar_unclustered(D, n, d):
    """Number of unclustered nodes at level d reachable from a^n"""
    if d == 0:
        return int(n == d)
    elif d <= D:
        return (d+1)*l_grammar_clustered(D,n,d) + l_grammar_clustered(D, n, d-1)
    elif d == D+1:
        return l_grammar_clustered(D,n,D)
    else:
        return 0

def l_complete_binary(n,d):
    """Number of nodes at level d reachable from a^n in complete binary tree"""
    if d >= n:
        return 2**(d-n)
    else:
        return 0

def expected_dfs(D, p, L):
    """For complete binary tree model"""
    if max(p) == 1: return np.argmax(p)+1
    # level rates
    l = [ -np.log(1-pk) for pk in p ]
    # total rate
    tr = sum([ lk/(2**(D-k+1)) for (k,lk) in enumerate(l) ])
    return 1/max(tr, 2**(-D))

def plot(D, pv, L, tv):
    res = np.zeros((len(pv), len(tv)))
    for (k, p) in enumerate(pv):
        print(p)
        res[k,:] = [ t(D, p, L) for t in tv ]
    plt.plot(res)
    plt.show()
    return res

# D = 15
# pv = np.zeros((D+1, D+1))
# for k in range(2,D+1):
#     pv[k,k] = 0.001 #1/2**(k-1)


assert l_grammar_unclustered(2, 0, 3) == 4
assert l_grammar_unclustered(2, 0, 2) == 14
assert l_grammar_unclustered(2, 0, 1) == 5
assert l_grammar_unclustered(2, 0, 0) == 1

assert l_grammar_unclustered(2, 1, 3) == 3
assert l_grammar_unclustered(2, 1, 2) == 10
assert l_grammar_unclustered(2, 1, 1) == 2
assert l_grammar_unclustered(2, 1, 0) == 0

assert l_grammar_unclustered(2, 2, 3) == 1
assert l_grammar_unclustered(2, 2, 2) == 3
assert l_grammar_unclustered(2, 2, 1) == 0
assert l_grammar_unclustered(2, 2, 0) == 0
