#
# Author:  Tom Everitt
# Email:   tom4everitt@gmail.com
# website: http://tomeveritt.se
#
# License:
# Use as you like as long as you credit me in publications.
#

import re
import csv
import math
import numpy as np
import random
import pandas
import matplotlib.pyplot as plt
from collections import deque
import graph_tool as gt
from graph_tool import draw, generation, search, stats, topology
from argparse   import ArgumentParser
import grammar
import safe_experiments

####################################################
#
# Helper methods
#
####################################################

def prod(list):
    """Return the product of elements in received list of numbers"""
    p = 1
    for n in list:
        p = p*n
    return p

def expectation_truncated_geometric(p, m):
    """Let X ~ Geo(p), i.e. P(X=k)=(1-p)^(k-1)*p. Return E[ X | X <= m]."""
    if p < 0.00000001: return m/2 + 1 # approximate to avoid division by zero
    return (1 - (1-p)**m*(p*m+1) )/(p*(1-(1-p)**m))

def sample_k(max):
    """Sample from cut-off 1/k distribution"""
    accept = False
    while not accept:
        k = random.randint(1,max+1)
        accept = random.random() < 1.0/k
    return k

def hist2mean(hist):
    """Compute mean from histogram on form (values, bucket_means)"""
    (values, bucket_means) = hist
    s = 0.0
    for i in range(len(values)):
        s += values[i]*bucket_means[i]
    return s/sum(values)

def hist2stdev(hist):
    """Compute stdev from histogram on form (values, bucket_means)"""
    (values, bucket_means) = hist
    mean = hist2mean(hist)
    sum_of_squares = 0
    for i in range(len(values)):
        sum_of_squares += values[i]*(bucket_means[i]**2)
    variance = sum_of_squares/sum(values) - mean**2
    return(math.sqrt(variance))

def random_string(alphabet, length):
    """Return random string from alphabet"""
    return ''.join([random.choice(alphabet) for i in range(length)])

def gamma(p, g):
    """gamma term in SGL decision boundary"""
    return np.log2(expectation_truncated_geometric(p, 2**g)-1)/2


#################################################
#
# Search algorithms
#
#################################################

def bfs_search(graph):
    """Return the number of BFS visited nodes until goal found"""
    
    class MyBFSVisitor(search.BFSVisitor):
        discovered_counter = 0
        found_goal = False
        def discover_vertex(self, v):
            self.discovered_counter += 1
            #print("discovered",v)
            if graph.goal[v]:
                self.found_goal = True
                #print("found goal", v,"in", self.discovered_counter, "steps")
                raise search.StopSearch()
    
    visitor = MyBFSVisitor()
    search.bfs_search(graph, graph.root, visitor=visitor)
    if visitor.found_goal:
        return visitor.discovered_counter
    else: # Search failed
        #return visitor.discovered_counter + 1
        raise NameError('No goal reached')

    
def dfs_search(graph):
    """Return the number of DFS visited nodes until goal found"""
    
    class MyDFSVisitor(search.DFSVisitor):
        discovered_counter = 0
        found_goal = False
        def discover_vertex(self, v):
            # print(v)
            self.discovered_counter += 1
            #print("discovered",v)
            if graph.goal[v]:
                self.found_goal = True
                #print("found goal", v,"in", self.discovered_counter, "steps")
                raise search.StopSearch()
    
    visitor = MyDFSVisitor()
    search.dfs_search(graph, graph.root, visitor=visitor)
    if visitor.found_goal:
        return visitor.discovered_counter
    else: # Search failed
        #return visitor.discovered_counter + 1
        raise NameError('No goal reached')

def my_dfs_rec(graph):
    """Alternative DFS implementation.
    Explore in same order as dfs_search"""
    discovered = {graph.root.__int__()}
    def rec(u, discovered):
        discovered.add(u.__int__())
        print(discovered)
        if graph.goal[u]: return len(discovered)
        for v in u.out_neighbours():
            print("considering", v)
            if not v.__int__() in discovered:
                c = rec(v, discovered)
                if c: return c
    return rec(graph.root, discovered) 
    
def my_dfs(graph):
    """Alternative DFS implementation.
    Explore in different order than dfs_search"""
    stack = [ graph.root ]
    discovered = { graph.root.__int__() } # set of discovered nodes
    counter = 0
    while len(stack) > 0:
        print(discovered)
        counter += 1
        u = stack.pop()
        if graph.goal[u]: return counter
        for v in u.out_neighbours():
            if not v.__int__() in discovered:
                if graph.goal[v]: return counter
                discovered.add(v.__int__())
                stack.append(v)
                
def my_bfs(graph):
    """Alternative BFS implementation"""
    queue = deque([ graph.root ])
    discovered = { graph.root } # set of discovered nodes
    counter = 0
    while len(queue) > 0:
        print(discovered)
        counter += 1
        u = queue.popleft()
        if graph.goal[u]: return counter
        for v in u.out_neighbours():
            if not v in discovered:
                discovered.add(v)
                queue.append(v)
            

    

#############################################
#
# Generic graph classes
#
#############################################

class GoalGraph(gt.Graph):
    """Generic graph with goals"""

    def __init__(self, directed=False):
        gt.Graph.__init__(self, directed=directed)
        self.root = self.add_vertex()
        self.goal = self.new_vertex_property("bool")
        self.goals = list()
    
    def make_goal(self, v):
        if not self.goal[v]:
            self.goal[v] = True
            self.goals.append(v)
    
    # Feature Methods
    
    def number_of_goals(self):
        return len(self.goals)
    
    def goal_frequency(self):
        return self.number_of_goals()/self.num_vertices()
    
    def pseudo_diameter(self):
        return topology.pseudo_diameter(self)[0]

    def shallowest_goal(self): 
        if len(self.goals) == 0: raise NameError('No goal')
        m = min([topology.shortest_distance(self, self.root, v, 
                                            max_dist=self.num_vertices()+1)
                 for v in self.goals])
        if m > self.num_vertices(): 
            raise NameError('No goal reachable') 
        else: return m

    def degree_histogram(self):
        return stats.vertex_hist(self, "out")
            
    def degree_mean(self):
        return hist2mean(self.degree_histogram())

    def degree_stdev(self):
        return hist2stdev(self.degree_histogram())
    
    def bfs_beats_dfs(self):
        return dfs_search(self) - bfs_search(self) > 0
    
    def bfs_wins(self):
        if dfs_search(self) > bfs_search(self):
            return "True"
        elif dfs_search(self) == bfs_search(self):
            return "Tie"
        else:
            return "False"

    def dfs_minus_bfs(self):
        return dfs_search(self) - bfs_search(self)
        
    def plotGraph(self, text=None):
        text = text or self.vertex_index
        pos = gt.draw.arf_layout(self)
        pos[self.vertex(0)] = [0.1,0.1]
        pos = gt.draw.arf_layout(self, pos=pos)
        color = self.new_vertex_property("string")
        for v in self.vertices():
            if self.goal[v]:
                color[v] = 'g'
            else:
                color[v] = 'w'
        color[self.root] = 'b'
        draw.graph_draw(self, pos, vertex_color=color, 
                        vertex_fill_color=color, 
                        vertex_text=text, 
                        vertex_font_size=10,
                        output_size=(2000, 2000)
                    )

        
class CompleteTree(GoalGraph):
    """Generic class for complete tree goal graphs"""
    
    def __init__(self, goal_prob=[0,0,0.15,0.1,0], branching_factor=2, create_graph=True):
        GoalGraph.__init__(self)
        self.D = len(goal_prob)-1 
        self.goal_prob_vec = goal_prob

        def add_children(v, d, max_d):
            if goal_prob[d] > random.random(): self.make_goal(v)
            if d == max_d: return # add no more children
            c_list = self.add_vertex(branching_factor) # add two chilren
            for c in c_list:
                self.add_edge(v,c)
                add_children(c, d+1, max_d)
        
        if create_graph: add_children(self.root, 0, self.D)
    
    def depth(self):
        return self.D

    def expected_bfs(self):
        """Expected BFS search time in a complete tree"""
        p = list(self.goal_prob_vec)
        p.append(1)
        f = [ self.level_has_first_goal(k) for k in range(len(p)) ]
        return sum([ f[k]*(2**k - 1 + expectation_truncated_geometric(pk, 2**k))
                     for k,pk in enumerate(p) if pk != 0 ])

    def expected_bfs_g(self):
        return self.expected_bfs() - (1-self.has_goal())*2**(self.D+1)
        
    def level_has_goal(self,k):
        """The probability that level k has a goal"""
        p = self.goal_prob_vec
        if k < len(p):
            return 1-(1-p[k])**(2**k)
        else:
            return 1
    
    def has_goal(self):
        """The probability that a goal exists"""
        p = self.goal_prob_vec
        return 1 - prod([(1 - self.level_has_goal(k)) for k,pk in enumerate(p)])

    def level_has_first_goal(self, k):
        """The probability that level k has the first goal"""
        p = self.goal_prob_vec
        return self.level_has_goal(k)*prod([(1-self.level_has_goal(j)) for j in range(k)])
    
    def approx_expected_bfs(self):
        """Approximation of the more complicated formula for expected_bfs"""
        p = list(self.goal_prob_vec)
        p.append(1)
        return sum([ self.level_has_first_goal(k)*(2**k + 1/pk) for k,pk in enumerate(p) if pk != 0])
    
    def expected_dfs(self):
        p = self.goal_prob_vec
        #if max(p) == 1: return np.argmax(p)+1
        D = self.D
        # level rates
        l = [ -math.log(1-pk) for pk in p ]
        # total rate
        tr = sum([ lk/(2**(D-k+1)) for (k,lk) in enumerate(l) ])
        return 1/max(tr, 2**(-D))

    def predicted_difference(self):
        """DFS-BFS. Negative means DFS better"""
        return self.expected_dfs()-self.expected_bfs()
    
    def predicted_difference_g(self):
        """DFS-BFS. Negative means DFS better"""
        return self.expected_dfs()-self.expected_bfs_g()
    
    def predicted_log_ratio(self):
        """log(DFS/BFS). Negative means DFS better"""
        return np.log(self.expected_dfs()/self.expected_bfs())
    
    def prediction_correct(self):
        return ((self.expected_dfs() - self.expected_bfs())*\
               (dfs_search(self)    - bfs_search(self)) > 0)*1.0

    def bfs_better_theory(self):
        return self.expected_dfs() > self.expected_bfs()
    
    feature_list = [
        depth,
        GoalGraph.shallowest_goal,
        expected_bfs,
        #expected_dfs,
        bfs_search
    ]
    
    @staticmethod
    def random_init():
        depth = random.randint(5,7)
        peak = depth-random.randint(1,4)
        interval = random.randint(1,4)
        goal_prob = [ (peak-interval < i < peak+interval)*0.03/depth 
                      for i in range(depth+1)]
        return {
            depth     : depth,
            goal_prob : goal_prob
        }


################################################
#
# Models investigated in Tree Search Paper
#
################################################

    
class CompleteTreeSGL(CompleteTree):
    """Single goal level model"""
    
    def __init__(self, depth=5, goal_depth=3, goal_prob=0.15, create_graph=True):
        assert int(depth) and int(goal_depth) and float(goal_prob)
        assert depth >= goal_depth >= 0
        assert 0 <= goal_prob <= 1
        goal_prob_vec = [ 0 for i in range(depth+1)]
        goal_prob_vec[math.ceil(goal_depth)] = goal_prob
        CompleteTree.__init__(self, goal_prob=goal_prob_vec, create_graph=create_graph)
        self.K = goal_depth
        self.p = goal_prob
    
    def goal_depth(self):
        return self.K
        
    def goal_depth_quote(self):
        return self.K/self.D
    
    def goal_prob(self):
        return self.p
                
    def bfs_better(self): # as predicted by theory
        #gamma = 1 + np.log2((1-self.p)/self.p)
        gammap = gamma(self.p, self.K)
        K = self.K
        D = self.D
        return (K < D/2 + gammap/2)

    def sgl_expected_bfs(self):
        G = self.has_goal()
        return G*(2**self.K -1 + expectation_truncated_geometric(self.p, 2**self.K)) + \
            (1-G)*2**(self.D+1)
    
    def sgl_expected_dfs(self):
        G = self.has_goal()
        return G*((expectation_truncated_geometric(self.p, 2**self.K) - 1)*2**(self.D-self.K+1) + 2) +\
            (1-G)*2**(self.D+1)
    
    def marcus_exp_dfs(self):
        return (1/self.p - 1)*2**(self.D-self.K+1) + 2
    
    feature_list = [
        bfs_search,
        dfs_search,
        #number_of_goals,
        CompleteTree.depth,
        CompleteTree.expected_dfs,
        #marcus_exp_dfs,
        #goal_depth,
        #goal_prob,
        #CompleteTree.expected_bfs,
        #CompleteTreeSGL.goal_depth,
        lambda g: int(bfs_search(g)<dfs_search(g)),
        #bfs_better
    ]
    
    @staticmethod
    def random_init():
        depth = random.randint(4, 15)
        goal_depth = random.randint(3,depth)
        #goal_prob = random.uniform(0.01, 0.5)
        #depth = 16
        #goal_depth = 12
        #goal_depth=depth-3
        goal_prob = 0.07
        return {
            'depth'      : depth,
            'goal_depth' : goal_depth,
            'goal_prob'  : goal_prob
        }


class CompleteTreeMGL(CompleteTree):
    """Gaussian multi-goal level model"""
    
    def __init__(self, depth=5, peak=3, spread=1, create_graph=True):
        self.K = peak
        self.s = spread
        goal_prob = [ min(np.exp( -(peak-k)**2/spread )*0.05/np.sqrt(spread), 1/2)
                      for k in range(depth+1) ]
        CompleteTree.__init__(self, goal_prob=goal_prob, create_graph=create_graph)
    
    def goal_peak(self):
        return self.K

    def goal_spread(self):
        return self.s
    
    @staticmethod
    def random_init():
        depth  = random.randint(7,15)
        middle = math.floor(depth/2)
        peak   = random.randint(middle-3, middle+3)
        return {
            depth    : depth,
            peak     : peak,
            spread   : 1
        }
    
    feature_list = [
        CompleteTree.depth,
        #goal_peak,
        #goal_spread,
        #max_prob,
        CompleteTree.expected_bfs,
        bfs_search,
        CompleteTree.expected_dfs,
        dfs_search,
        CompleteTree.prediction_correct
    ]


####################################################
#
# Grammar graphs
#
####################################################

class Grammar(GoalGraph):
    """Generic grammar graph"""

    def __init__(self, 
                 start_string = "S", 
                 goal_strings = ["aaa"], 
                 rules        = [('S', ''), ('S','Sa'), ('S','Sb')], 
                 max_depth    = 3, # negative = infinity
                 create_graph = True
                ):
        GoalGraph.__init__(self, directed=True)
        self.string       = self.new_vertex_property("string")
        self.node         = {} # dictionary string->node
        self.goal_strings = goal_strings
        self.rules        = rules
        self.max_depth    = max_depth
        
        def add_children(v, max_d):
            s = self.string[v]
            if s in goal_strings: self.make_goal(v)
            if len(self.remove_S(s)) == max_d: 
                loc_rules = [('S', '')]
            else: loc_rules = rules
            for r in loc_rules:
                new_strings = self.apply_rule(r, s)
                for new_s in new_strings:
                    if new_s in self.node:
                        # node already exists, just add a link to it
                        self.add_edge(v, self.node[new_s])
                    else:
                        # create a node and recursively add its children
                        child = self.add_node(new_s)
                        self.add_edge(v, child)
                        add_children(child, max_d)
                        
        self.string[self.root]  = start_string
        self.node[start_string] = self.root
        if create_graph:
            add_children(self.root, max_depth)
        
    def remove_S(self, string):
        """Return string with S removed if occurs"""
        ns = self.apply_rule(('S', ''), string)
        if len(ns)>0: return ns[0]
        return string
    
    def apply_rule(self, rule, string):
        """Return the list of strings creatable from string and rule"""
        (o, n) = rule
        subs = [(i.start(), i.end()) for i in re.finditer(o, string)]
        new_s = set()
        for (a,b) in subs:
            new_s.add(string[:a] + n + string[b:])
        return list(new_s)

    def add_node(self, string):
        if string in self.node: # already added
            return self.node[string]
        else:
            v = self.add_vertex()
            self.string[v]    = string
            self.node[string] = v
            return v
    
    def number_of_rules(self):
        return len(self.rules)
    
    def maximum_depth(self):
        return self.max_depth

    def num_goal_strings(self):
        return len(self.goal_strings)

    def mean_len_goal_str(self):
        return np.mean([len(s) for s in self.goal_strings])

    def stdev_len_goal_str(self):
        return np.std([len(s) for s in self.goal_strings])
    
    def grammar_power(self):
        return self.number_of_goals()/self.num_goal_strings()
    
    def plotGraph(self):
        GoalGraph.plotGraph(self, text=self.string)



class BinaryGrammar(Grammar):
    """Model investigated in paper as Binary Grammar"""
    
    def __init__(self, depth = 3, g = 2, pg=0.25, create_graph=True):
        self.depth = depth
        self.g = g
        self.pg = pg
        rules = [ ('', 'a'), ('', 'b') ]
        start_string = ''
        # This sampling of goals is equivalent to iid sampling with prob p
        num_goals = np.random.binomial(2**g, pg)
        print("num_goals", num_goals)
        goals = [ random_string(['a', 'b'], g) for j in range(num_goals) ]
        Grammar.__init__(self, rules=rules, max_depth=depth, 
                         start_string=start_string, goal_strings=goals, 
                         create_graph=create_graph)
    
    def expected_dfs(self):
        return ( grammar.tdfs_l_clust(self.depth, self.g, self.pg) + 
                 grammar.tdfs_u_clust(self.depth, self.g, self.pg) )/2
        
    def expected_dfsl(self):
        return grammar.tdfs_l_clust(self.depth, self.g, self.pg)
                 
    def expected_dfsu(self):
        return grammar.tdfs_u_clust(self.depth, self.g, self.pg)

    def expected_bfs(self):
        return grammar.tbfs_clust(self.depth, self.g, self.pg)

    def goal_depth(self):
        return self.g
    
    def goal_prob(self):
        return self.pg

    def predicted_difference(self):
        return self.expected_dfs() - self.expected_bfs()

    def bfs_better_theory(self):
        return self.expected_dfs() > self.expected_bfs()

    feature_list = [
        bfs_search,
        dfs_search,
        expected_dfs,
        expected_dfsl,
        expected_dfsu,
        expected_bfs,
    ]

    
class CompleteGrammarTree(Grammar):
    """Grammar that creates a complete tree for crosschecking"""
    def __init__(self, goal_prob=[0,0,0.5]):
        rules = [('S', 'Sa'), ('S', 'Sb')] 
        goal_strings = []
        for i,p in enumerate(goal_prob):
            for k in range(2**i, 2**(i+1)):
                if random.random() < p:
                    c  = {'0': 'a', '1': 'b'}
                    s  = "S"
                    kb = '{0:b}'.format(k)
                    for i in kb:
                        s = s + c[i]
                    goal_strings.append(s)
        Grammar.__init__(self, 
                         start_string = 'S', 
                         rules        = rules, 
                         goal_strings = goal_strings, 
                         max_depth    = len(goal_prob))

        
############################################
#
# Random graphs
#
############################################

class RandomGrammar(Grammar):

    def __init__(self, 
                 num_rules        = 3,  # number of rules in addition to S->''
                 num_goal_strings = 3,  # number of goal strings
                 len_goal_strings = lambda: random.randint(1, 3), #length of goal strings
                 max_depth        = 4):  # maximum depth of tree
        rules = [ ('S', '') ]
        possible_rules = [
            ('S', 'Sa'), 
            ('S', 'Sb'), 
            ('S', 'aS'), 
            ('S', 'bS'),
            ('Sa', 'aS'),
            ('Sb', 'bS'),
            ('aS', 'Sa'),
            ('bS', 'Sb'),
        ]
        rules.extend(random.sample(possible_rules, num_rules))
        goals = [ random_string(['a', 'b'], len_goal_strings()) 
                  for j in range(num_goal_strings) ]
        Grammar.__init__(self, 
                         start_string = 'S', 
                         goal_strings = goals, 
                         rules        = rules, 
                         max_depth    = max_depth)
    
    @staticmethod
    def random_init():
        max_depth        = 15 #random.randint(11,15)
        num_goal_strings = random.randint(3, 5) # *max_depth)
        lgl              = 2 #random.randint(1, 3)
        ugl              = 5 #random.randint(lgl, 5) # max_depth)
        len_goal_strings = lambda: random.randint(lgl, ugl)
        num_rules        = random.randint(3, 8)
        return dict(
            max_depth        = max_depth,
            num_goal_strings = num_goal_strings,
            len_goal_strings = len_goal_strings,
            num_rules        = num_rules
        )

    feature_list = [
        #GoalGraph.degree_mean,
        #GoalGraph.degree_stdev,
        #Grammar.number_of_rules,
        #Grammar.maximum_depth,
        #Grammar.num_goal_strings,
        #Grammar.mean_len_goal_str,
        #Grammar.stdev_len_goal_str,
        # Questionable features:
        #Grammar.grammar_power,
        GoalGraph.number_of_goals,
        #GoalGraph.shallowest_goal,
        #GoalGraph.pseudo_diameter,
        #GoalGraph.goal_frequency,
        bfs_search,
        dfs_search,
        #GoalGraph.bfs_beats_dfs,
        #GoalGraph.dfs_minus_bfs,
    ]

class RandomGraph(GoalGraph):
    """Create a goal graph with random structure and 
    randomly sampled goals at specified distance from root"""

    def __init__(self, 
                 N=40, 
                 goal_level=3, 
                 number_of_goals=2, 
                 degree=lambda: sample_k(5)):
        GoalGraph.__init__(self)
        self.K = goal_level
        self.number_of_goals = number_of_goals
        # Copy a random graph to self
        rgraph = generation.random_graph(N, degree, directed=False)
        self.add_vertex(N)
        for e in rgraph.edges():
            self.add_edge(e.source().__int__(), e.target().__int__())
        # Put goals at the right level (rejection sampling, inefficient)
        # could be made more efficient by BFS layering the graph
        g = 0
        i = 0
        while g < self.number_of_goals and i < 10*N:
            i += 1
            v = self.vertex(random.randint(1, N))
            if topology.shortest_distance(self, self.root, v)==goal_level and \
               not self.goal[v]:
                self.goal[v] = True
                self.goals.append(v)
                g += 1        
        if i >= 10*N:
            print("goaling failed")
    
    @staticmethod
    def random_init():
        N=random.randint(10,100)
        goal_level=random.randint(3,max(3,int(N/10)))
        number_of_goals=random.randint(1,goal_level)
        degree=lambda: sample_k(5)
        return {
            N               : N, 
            goal_level      : goal_level, 
            number_of_goals : number_of_goals, 
            degree          : degree
        }

    # feature_list = [
    #     number_of_goals,
    #     pseudo_diameter,
    #     Goalshallowest_goal,
    #     degree_mean,
    #     degree_stdev,
    #     bfs_search,
    #     dfs_search,
    #     lambda g: bfs_search(g)-dfs_search(g)
    # ]

    

#################################################
#
# Experiment methods
#
#################################################

def add_pd_row(df, row):
    df.loc[df.shape()[0]] = row

def getcsv(graph_class, features = None, outfile = None):
    filename = outfile or graph_class.__name__ + ".csv"
    features = features or graph_class.feature_list
    #print("features",features)
    csvfile = csv.writer(sf.open(filename), delimiter='\t')
    csvfile.writerow([f.__name__ for f in features])
    return csvfile

def experiment(graph_class, features=None, 
               random_init=None, 
               csvfile = None, # csv.writer(open(filename, 'a', 1))
               outfile = None,
               append = False,
               iterate=10):
    """Run experiment with given graph class, and output piped to
    given file"""
    outfile = outfile or graph_class.__name__ + ".csv"
    features = features or graph_class.feature_list
    random_init = random_init or graph_class.random_init
    if append and not csvfile:
        csvfile = csv.writer(sf.open(outfile, mode='a'), delimiter='\t')
    else: 
          try:  
              csvfile = csvfile or getcsv(graph_class, features, outfile = outfile)
          except FileExistsError as err:
              print("file exists. Remove first or use -a to append to it.")
              return
    rows_list = np.zeros((0,len(features)))
    bfs_win = 0
    print(iterate)
    # TODO: Multi-thread?
    for i in range(iterate):
        g = graph_class(**random_init())
        try:
            feats = [f(g) for f in features]
            rows_list = np.append(rows_list, np.array([feats]), axis=0)
            csvfile.writerow(feats)
            print(i, end=" ", flush=True)
        except NameError as err:
            print('X', end=" ", flush=True)
    df = pandas.DataFrame(rows_list, columns=[f.__name__ for f in features])
    return df

def averages(df):
    return pandas.DataFrame([[round(np.average(df[c]), 2) for c in df]], columns=df.columns.values)

def correlations(df):
    return np.round(np.corrcoef([df[c] for c in df]), 2)

def perc_err(correct, guess):
    """The error of the guess in percentage"""
    return 100*np.abs(correct-guess)/correct

##########################################################
#
# Methods to verify analytical approximations.
# (Uncomment calls to run.)
#
##########################################################


def verifySGL(krange, prange, D=15, iterate=1000):
    feature_list = [
        CompleteTreeSGL.sgl_expected_dfs,
        dfs_search,
        CompleteTreeSGL.sgl_expected_bfs,
        bfs_search,
    ]
    dfsana = np.zeros((len(krange), len(prange)))
    dfsexp = np.zeros((len(krange), len(prange)))
    dfserr = np.zeros((len(krange), len(prange)))    

    bfsana = np.zeros((len(krange), len(prange)))
    bfsexp = np.zeros((len(krange), len(prange)))
    bfserr = np.zeros((len(krange), len(prange)))
    for (ki, k) in enumerate(krange):
        print('k', k)
        for (pi, p) in enumerate(prange):
            #if p < 2**(-k-2): continue 
            print('k p', k, p)
            random_init = lambda: dict(depth=D, goal_depth=k, goal_prob=p)
            df = experiment(CompleteTreeSGL, 
                            features=feature_list, 
                            outfile = 'SGL_D'+str(D)+'p'+str(p)+'k'+str(k),
                            random_init=random_init, iterate=iterate, append=True)
            av = averages(df)
            dfsana[ki,pi] = av['sgl_expected_dfs']
            dfsexp[ki,pi] = av['dfs_search']
            dfserr[ki,pi] = perc_err(av['dfs_search'], av['sgl_expected_dfs'])
            bfsana[ki,pi] = av['sgl_expected_bfs']
            bfsexp[ki,pi] = av['bfs_search']
            bfserr[ki,pi] = perc_err(av['bfs_search'], av['sgl_expected_bfs'])

            print('av\n', av)
    return ( [bfsexp, bfsana, bfserr], [dfsexp, dfsana, dfserr] )

# Verify SGL
# sf = safe_experiments.SafeFiles('sgl')
# D = 14
# krange = [5, 8, 11, 14]
# prange = [0.001, 0.003, 0.01, 0.03, 0.1, 0.3]
# verifySGL(krange, prange, D=D, iterate=10)


def verifyMGL(krange, srange, D=15, iterate=1000):
    feature_list = [
        CompleteTree.expected_dfs,
        dfs_search,
        CompleteTree.expected_bfs_g,
        bfs_search,
    ]
    dfsana = np.zeros((len(krange), len(srange)))
    dfsexp = np.zeros((len(krange), len(srange)))
    dfserr = np.zeros((len(krange), len(srange)))    

    bfsana = np.zeros((len(krange), len(srange)))
    bfsexp = np.zeros((len(krange), len(srange)))
    bfserr = np.zeros((len(krange), len(srange)))
    for (ki, k) in enumerate(krange):
        print('k', k)
        for (si, s) in enumerate(srange):
            #if p < 2**(-k-2): continue 
            print('k s', k, s)
            random_init = lambda: dict(depth=D, peak=k, spread=s)
            df = experiment(CompleteTreeMGL, 
                            features=feature_list, 
                            outfile = 'MGL_D'+str(D)+'k'+str(k)+'s'+str(s),
                            random_init=random_init, 
                            iterate=iterate, append=True)
            av = averages(df)
            dfsana[ki,si] = av['expected_dfs']
            dfsexp[ki,si] = av['dfs_search']
            dfserr[ki,si] = perc_err(av['dfs_search'], av['expected_dfs'])
            bfsana[ki,si] = av['expected_bfs_g']
            bfsexp[ki,si] = av['bfs_search']
            bfserr[ki,si] = perc_err(av['bfs_search'], av['expected_bfs_g'])

            print('av\n', av)
    return ( [bfsexp, bfsana, bfserr], [dfsexp, dfsana, dfserr] )

# Verify MGL
# sf = safe_experiments.SafeFiles('mgl')
# D = 14
# krange = [5, 8, 11, 14]
# srange = [0.1, 1, 10, 100]
# verifyMGL(krange, srange, D=D, iterate=10)




def verifyCG(krange, prange, D=15, iterate=1000):
    feature_list = [
        BinaryGrammar.expected_dfsl,
        dfs_search,
        BinaryGrammar.expected_dfsu,
        BinaryGrammar.expected_dfs,
        BinaryGrammar.expected_bfs,
        bfs_search,
    ]
    ldfsana = np.zeros((len(krange), len(prange)))
    ldfsexp = np.zeros((len(krange), len(prange)))
    ldfserr = np.zeros((len(krange), len(prange)))

    udfsana = np.zeros((len(krange), len(prange)))
    udfsexp = np.zeros((len(krange), len(prange)))
    udfserr = np.zeros((len(krange), len(prange)))       

    dfsana = np.zeros((len(krange), len(prange)))
    dfsexp = np.zeros((len(krange), len(prange)))
    dfserr = np.zeros((len(krange), len(prange)))

    bfsana = np.zeros((len(krange), len(prange)))
    bfsexp = np.zeros((len(krange), len(prange)))
    bfserr = np.zeros((len(krange), len(prange)))
    for (ki, k) in enumerate(krange):
        print('k', k)
        for (pi, p) in enumerate(prange):
            print('k p', k, p)
            random_init = lambda: dict(depth=D, g=k, pg=p)
            df = experiment(BinaryGrammar, 
                            features=feature_list, 
                            outfile = 'SGL_D'+str(D)+'p'+str(p)+'k'+str(k),
                            random_init=random_init, iterate=iterate, append=True)
            av = averages(df)
            ldfsana[ki,pi] = av['expected_dfsl']
            ldfsexp[ki,pi] = av['dfs_search']
            ldfserr[ki,pi] = perc_err(av['dfs_search'], av['expected_dfsl'])
            udfsana[ki,pi] = av['expected_dfsu']
            udfsexp[ki,pi] = av['dfs_search']
            udfserr[ki,pi] = perc_err(av['dfs_search'], av['expected_dfsu'])
            dfsana[ki,pi] = av['expected_dfs']
            dfsexp[ki,pi] = av['dfs_search']
            dfserr[ki,pi] = perc_err(av['dfs_search'], av['expected_dfs'])
            bfsana[ki,pi] = av['expected_bfs']
            bfsexp[ki,pi] = av['bfs_search']
            bfserr[ki,pi] = perc_err(av['bfs_search'], av['expected_bfs'])

            print('av\n', av)
    return ( [bfsexp, bfsana, bfserr], [ldfsexp, ldfsana, ldfserr], [udfsexp, udfsana, udfserr], [dfsexp, dfsana, dfserr] )

# Verify CG
# sf = safe_experiments.SafeFiles('cg')
# D = 14
# krange = [5, 8, 11, 14]
# prange = [0.001, 0.01, 0.1]
# verifyCG(krange, prange, D=D, iterate=10)



##################################################
#
# If called as script
# Example call: python3 -i 10 CompleteTreeSGL
# Problems are generated according to the class's random_init method,
# and run with the methods in the class's feature_list
#
##################################################

if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("class_name")
    parser.add_argument("-i", "--iterate", type = int)
    parser.add_argument("-o", "--outfile")
    parser.add_argument("-f", "--features")
    parser.add_argument("-a", "--append",  action="store_true")
    args = parser.parse_args()
    class_name = eval(args.class_name)
    features = eval(args.features or 'None')
    sf = safe_experiments.SafeFiles(args.class_name)
    i = args.iterate or 10
    experiment(class_name, features=features, iterate=i,
               outfile=args.outfile, append=args.append)

